import math
from typing import Literal

import torch
import torch.nn.functional as F
from torch import nn

from eir.models.fusion.fusion_attention import UniDirectionalCrossAttention
from eir.models.layers.cnn_layers import adjust_num_heads
from eir.models.layers.norm_layers import LayerScale
from eir.models.tensor_broker.projection_modules.sequence import (
    get_reshape_to_attention_dims_func,
)


class ProjectAndFuseLayer(nn.Module):
    def __init__(
        self,
        projection_layer: nn.Module,
        fusion_layer: nn.Module,
    ):
        super().__init__()
        self.projection_layer = projection_layer
        self.fusion_layer = fusion_layer

    def forward(
        self,
        input_tensor: torch.Tensor,
        cached_tensor: torch.Tensor,
    ) -> torch.Tensor:
        projected = self.projection_layer(cached_tensor)
        fused = self.fusion_layer(input_tensor, projected)
        return fused


def get_fusion_layer_wrapper(
    projected_shape: torch.Size,
    target_shape: torch.Size,
    cache_fusion_type: Literal["cross-attention", "sum", "cat+conv"],
    projection_layer: nn.Module,
    device: str,
) -> ProjectAndFuseLayer:
    fusion_layer = get_fusion_layer(
        projected_shape=projected_shape,
        target_shape=target_shape,
        cache_fusion_type=cache_fusion_type,
    )
    project_and_fuse_layer = ProjectAndFuseLayer(
        projection_layer=projection_layer,
        fusion_layer=fusion_layer,
    )
    project_and_fuse_layer = project_and_fuse_layer.to(device=device)
    return project_and_fuse_layer


def get_fusion_layer(
    target_shape: torch.Size,
    projected_shape: torch.Size,
    cache_fusion_type: Literal["cross-attention", "sum", "cat+conv"],
    projection_type: Literal["lcl", "lcl_residual", "cnn", "linear", "pool"] = "linear",
) -> nn.Module:

    match cache_fusion_type:
        case "cross-attention":
            return CrossAttentionFusionLayer(
                input_shape=target_shape,
                context_shape=projected_shape,
            )
        case "sum":
            return GatedSumFusionLayer(
                input_shape=target_shape,
                context_shape=projected_shape,
            )
        case "cat+conv":
            return ConcatenationFusionLayer(
                input_shape=target_shape,
                context_shape=projected_shape,
            )
        case _:
            raise ValueError(f"Invalid cache_fusion_type: {cache_fusion_type}")


class GatedSumFusionLayer(nn.Module):
    def __init__(
        self,
        input_shape: torch.Size,
        context_shape: torch.Size,
        initial_input_bias: float = 0.9,
    ):
        super().__init__()
        self.input_shape = input_shape
        self.context_shape = context_shape

        input_init = torch.Tensor([math.log(initial_input_bias)])
        projected_init = torch.Tensor([math.log(1 - initial_input_bias)])
        self.gate_input = nn.Parameter(input_init, requires_grad=True)
        self.gate_projected = nn.Parameter(projected_init, requires_grad=True)

    def forward(
        self, input_tensor: torch.Tensor, projected_context_tensor: torch.Tensor
    ) -> torch.Tensor:
        gates = F.softmax(torch.stack([self.gate_input, self.gate_projected]), dim=0)

        scaled_input = input_tensor * gates[0]
        scaled_projected = projected_context_tensor * gates[1]

        return scaled_input + scaled_projected


class ConcatenationFusionLayer(nn.Module):
    def __init__(
        self,
        input_shape: torch.Size,
        context_shape: torch.Size,
    ):
        """
        Here we assume that the context has already been projected to
        (batch_size, input_seq_length, input_embedding_dim) when calling forward.
        """
        super().__init__()
        self.context_shape = context_shape
        self.input_shape = input_shape

        self.context_channels = context_shape[0]
        self.out_channels = input_shape[0]

        cat_dim = self.context_channels + self.out_channels

        self.conv_1 = nn.Conv2d(
            in_channels=cat_dim,
            out_channels=self.out_channels,
            kernel_size=1,
        )
        self.norm = nn.GroupNorm(1, self.out_channels)
        self.act = nn.GELU()

        self.ls = LayerScale(dim=1, init_values=1e-05)

    def forward(
        self, input_tensor: torch.Tensor, projected_context_tensor: torch.Tensor
    ) -> torch.Tensor:
        residual = input_tensor
        out = torch.cat((input_tensor, projected_context_tensor), dim=1)
        out = self.conv_1(out)
        out = self.norm(out)
        out = self.act(out)
        out = self.ls(out)

        return out + residual


class CrossAttentionFusionLayer(nn.Module):
    def __init__(
        self,
        input_shape: torch.Size,
        context_shape: torch.Size,
    ):
        super().__init__()
        self.input_shape = input_shape
        self.context_shape = context_shape

        (
            self.reshape_func,
            self.reshaped_size,
        ) = get_reshape_to_attention_dims_func(
            input_shape=input_shape,
        )

        self.embedding_dim = self.reshaped_size[1]
        self.context_dim = context_shape[1]

        self.n_heads = adjust_num_heads(num_heads=8, embedding_dim=self.embedding_dim)

        self.norm_1 = nn.LayerNorm(self.embedding_dim)
        self.act_1 = nn.GELU()

        self.cross_attention = UniDirectionalCrossAttention(
            dim=self.embedding_dim,
            dim_head=self.embedding_dim // self.n_heads,
            context_dim=self.context_dim,
            heads=self.n_heads,
        )

        self.ls = LayerScale(dim=1, init_values=1e-5)

    def forward(
        self, input_tensor: torch.Tensor, projected_context_tensor: torch.Tensor
    ) -> torch.Tensor:
        identity = input_tensor

        out = self.reshape_func(input_tensor)
        out = self.norm_1(out)

        out = self.cross_attention(x=out, context=projected_context_tensor)
        out = self.act_1(out)
        out = self.ls(out)

        out = out.reshape(identity.shape)

        return out + identity



import atexit
import json
import threading
import time
from typing import TYPE_CHECKING, Any, Iterator, Literal, Optional, Sequence, Union

import numpy as np
import torch
import websocket
from PIL import Image
from sklearn.preprocessing import LabelEncoder, StandardScaler
from torch.utils.data import IterableDataset

from eir.data_load.data_preparation_modules.imputation import (
    impute_missing_modalities_wrapper,
    impute_missing_output_modalities_wrapper,
)
from eir.data_load.data_preparation_modules.input_preparation_wrappers import (
    prepare_inputs_memory,
)
from eir.data_load.data_preparation_modules.output_preparation_wrappers import (
    prepare_outputs_memory,
)
from eir.data_load.data_utils import Sample
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_network_utils import (
    deserialize_array,
    deserialize_image,
    prepare_request_input_data,
    streamline_sequence_manual_data,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.output_setup import ComputedArrayOutputInfo, al_output_objects_as_dict
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import ImageOutputTypeConfig, SequenceOutputTypeConfig
from eir.setup.streaming_data_setup.protocol import PROTOCOL_VERSION
from eir.setup.streaming_data_setup.streaming_data_utils import (
    connect_to_server,
    receive_with_timeout,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.datasets import al_getitem_return


al_outputs_prepared = dict[
    str,
    np.ndarray | torch.Tensor | list[str] | str | dict | Image.Image,
]

logger = get_logger(name=__name__)


class StreamingDataset(IterableDataset):
    def __init__(
        self,
        websocket_url: str,
        inputs: al_input_objects_as_dict,
        outputs: al_output_objects_as_dict,
        test_mode: bool,
        batch_size: int = 32,
        fetch_timeout: float = 10.0,
        max_consecutive_timeouts: int = 3,
        heartbeat_interval: float = 30.0,
    ):
        self.websocket_url = websocket_url
        self.inputs = inputs
        self.outputs = outputs
        self.test_mode = test_mode
        self.batch_size = batch_size
        self.fetch_timeout = fetch_timeout
        self.max_consecutive_timeouts = max_consecutive_timeouts
        self.heartbeat_interval = heartbeat_interval
        self.ws: Optional[websocket.WebSocket] = None
        self.current_batch: list[dict] = []
        self.heartbeat_thread: Optional[threading.Thread] = None
        self.stop_event = threading.Event()
        self._is_closed = False

        atexit.register(self.close)

    def __iter__(self) -> Iterator[Any]:
        self._connect()
        return self

    def __next__(self) -> Any:
        if not self.current_batch:
            self._fetch_batch()

        if not self.current_batch:
            raise StopIteration

        sample_dict = self.current_batch.pop(0)
        sample = Sample(**sample_dict)
        return self._process_sample(sample)

    def _connect(self):
        if self.heartbeat_thread and self.heartbeat_thread.is_alive():
            self.stop_event.set()
            self.heartbeat_thread.join(timeout=5)

        self.connection_context = connect_to_server(
            websocket_url=self.websocket_url,
            protocol_version=PROTOCOL_VERSION,
        )
        self.ws = self.connection_context.__enter__()
        self.stop_event.clear()
        self.heartbeat_thread = threading.Thread(
            target=self._send_heartbeats,
            daemon=True,
        )
        self.heartbeat_thread.start()

    def _send_heartbeats(self):
        while not self.stop_event.is_set():
            if self.ws:
                try:
                    self.ws.send(json.dumps({"type": "heartbeat"}))
                except Exception as e:
                    logger.error(f"Error sending heartbeat: {e}")
            time.sleep(self.heartbeat_interval)

    def _fetch_batch(self):
        consecutive_timeouts = 0
        assert self.ws is not None
        while consecutive_timeouts < self.max_consecutive_timeouts:
            try:
                self.ws.send(
                    json.dumps(
                        {
                            "type": "getData",
                            "payload": {"batch_size": self.batch_size},
                        }
                    )
                )

                batch_msg = receive_with_timeout(websocket=self.ws)

                if batch_msg["type"] == "heartbeat":
                    continue

                if batch_msg["type"] != "data":
                    logger.warning(f"Unexpected message type: {batch_msg['type']}")
                    continue

                if not batch_msg["payload"] or batch_msg["payload"] == ["terminate"]:
                    logger.info(
                        "Received termination signal or empty payload, stopping."
                    )
                    raise StopIteration

                self.current_batch = batch_msg["payload"]
                return

            except websocket.WebSocketTimeoutException:
                consecutive_timeouts += 1
                logger.warning(
                    f"Timeout waiting for next batch. "
                    f"Consecutive timeouts: {consecutive_timeouts}"
                )
            except websocket.WebSocketConnectionClosedException:
                logger.warning("WebSocket connection closed. Attempting to reconnect.")
                self._reconnect()
            except OSError as e:
                if e.errno == 32:
                    logger.warning("Broken pipe detected. Attempting to reconnect.")
                    self._reconnect()
                else:
                    logger.error(f"Unexpected OSError: {e}")
                    raise StopIteration
            except Exception as e:
                logger.error(f"Error fetching batch: {e}")
                raise StopIteration

        logger.error(
            f"Max consecutive timeouts ({self.max_consecutive_timeouts}) reached."
            f" Stopping iterator."
        )
        raise StopIteration

    def _reconnect(self):
        if self.ws:
            self.connection_context.__exit__(None, None, None)
        self._connect()

    def close(self):
        if not self._is_closed:
            self._is_closed = True
            self.stop_event.set()
            if self.ws:
                self.connection_context.__exit__(None, None, None)
            atexit.unregister(self.close)

    def _process_sample(self, sample: Sample) -> "al_getitem_return":

        inputs = sample.inputs
        inputs_request_parsed = prepare_request_input_data(
            request_data=inputs,
            input_objects=self.inputs,
        )

        inputs_prepared_for_memory = prepare_inputs_for_in_memory_processing(
            inputs_request_parsed=inputs_request_parsed,
            input_objects=self.inputs,
        )

        inputs_prepared = prepare_inputs_memory(
            inputs=inputs_prepared_for_memory,
            inputs_objects=self.inputs,
            test_mode=self.test_mode,
        )

        inputs_final = impute_missing_modalities_wrapper(
            inputs_values=inputs_prepared,
            inputs_objects=self.inputs,
        )

        target_labels = sample.target_labels

        targets_request_parsed = prepare_request_output_data(
            request_data=target_labels,
            output_objects=self.outputs,
        )

        targets_prepared_for_memory = prepare_outputs_for_in_memory_processing(
            target_labels=targets_request_parsed,
            output_objects=self.outputs,
        )

        targets_prepared = prepare_outputs_memory(
            outputs=targets_prepared_for_memory,
            output_objects=self.outputs,
            test_mode=self.test_mode,
        )

        targets_final = impute_missing_output_modalities_wrapper(
            outputs_values=targets_prepared,
            output_objects=self.outputs,
        )

        sample_id = sample.sample_id

        return inputs_final, targets_final, sample_id


def prepare_inputs_for_in_memory_processing(
    inputs_request_parsed: dict[str, Any],
    input_objects: al_input_objects_as_dict,
) -> dict[str, Any]:

    inputs_prepared_for_memory = {}
    for name, cur_input in inputs_request_parsed.items():
        input_object = input_objects[name]

        match input_object:
            case ComputedSequenceInputInfo():
                cur_input = input_object.encode_func(cur_input)

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                cur_input = _impute_missing_tabular_values(
                    input_object=input_object,
                    inputs_values=cur_input,
                )

        inputs_prepared_for_memory[name] = cur_input

    return inputs_prepared_for_memory


def prepare_request_output_data(
    request_data: dict[str, Any],
    output_objects: "al_output_objects_as_dict",
) -> dict[str, Any]:
    outputs_prepared: al_outputs_prepared = {}

    for output_name, serialized_data in request_data.items():

        output_object = output_objects[output_name]

        output_config = output_object.output_config
        output_type = output_config.output_info.output_type
        output_type_info = output_config.output_type_info

        match output_object:
            case ComputedSequenceOutputInfo():
                assert output_type == "sequence"
                assert isinstance(output_type_info, SequenceOutputTypeConfig)

                value = serialized_data[output_name]
                sequence_streamlined = streamline_sequence_manual_data(
                    data=value,
                    split_on=output_type_info.split_on,
                )

                outputs_prepared[output_name] = {output_name: sequence_streamlined}

            case ComputedImageOutputInfo():
                assert output_type == "image"
                assert isinstance(output_type_info, ImageOutputTypeConfig)

                value = serialized_data[output_name]
                image_data = deserialize_image(
                    image_str=value,
                    image_mode=output_type_info.mode,
                )

                outputs_prepared[output_name] = {output_name: image_data}

            case ComputedTabularOutputInfo():
                assert output_type == "tabular"
                tabular_data = apply_tabular_transformers(
                    cur_target=serialized_data,
                    target_transformers=output_object.target_transformers,
                    target_columns=output_object.target_columns,
                )

                outputs_prepared[output_name] = tabular_data

            case ComputedArrayOutputInfo():
                assert output_type == "array"

                value = serialized_data[output_name]
                array_np = deserialize_array(
                    array_str=value,
                    dtype=output_object.dtype,
                    shape=output_object.data_dimensions.full_shape(),
                )
                outputs_prepared[output_name] = {output_name: array_np}

            case _:
                raise ValueError(f"Unknown output type '{output_type}'")

    return outputs_prepared


def prepare_outputs_for_in_memory_processing(
    target_labels: dict[str, Any],
    output_objects: al_output_objects_as_dict,
) -> dict[str, Any]:

    targets_prepared_for_memory = {}
    for name, cur_target in target_labels.items():
        output_object = output_objects[name]

        match output_object:
            case ComputedSequenceOutputInfo():

                value = cur_target[name]
                assert isinstance(value, Sequence), value

                cur_target = output_object.encode_func(value)
                targets_prepared_for_memory[name] = {name: cur_target}

            case _:
                targets_prepared_for_memory[name] = cur_target

    return targets_prepared_for_memory


def apply_tabular_transformers(
    cur_target: dict[str, Union[int, float, torch.Tensor]],
    target_transformers: dict[str, Union[StandardScaler, LabelEncoder]],
    target_columns: dict[Literal["con", "cat"], list[str]],
) -> dict[str, int | float | torch.Tensor]:
    transformed_target = {}

    for col_type, columns in target_columns.items():
        for column in columns:
            if column not in cur_target:
                raise ValueError(f"Column {column} not found in target data.")

            if column not in target_transformers:
                raise ValueError(f"Transformer for column {column} not found.")

            transformer = target_transformers[column]
            value = cur_target[column]

            if isinstance(transformer, StandardScaler):
                transformed_value = transformer.transform([[value]])[0]
            elif isinstance(transformer, LabelEncoder):
                transformed_value = transformer.transform([value])
            else:
                raise ValueError(f"Unknown transformer type for column {column}")

            transformed_target[column] = transformed_value.item()

    return transformed_target


def _impute_missing_tabular_values(
    input_object: (
        ComputedTabularInputInfo
        | ComputedPredictTabularInputInfo
        | ComputedServeTabularInputInfo
    ),
    inputs_values: dict[str, Any],
) -> dict[str, Any]:
    # TODO: Implement
    return inputs_values



import os
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    overload,
)

import aislib.misc_utils
import numpy as np
import torch
import yaml
from ignite.contrib.handlers import ProgressBar
from ignite.engine import CallableEventWithFilter, Engine, Events, EventsList, events
from ignite.handlers import EarlyStopping, ModelCheckpoint
from ignite.metrics import RunningAverage
from ignite.metrics.metric import Metric, RunningBatchWise
from torch import nn

from eir.data_load.data_utils import get_output_info_generator
from eir.interpretation.interpretation import attribution_analysis_handler
from eir.setup.config_setup_modules.config_setup_utils import object_to_primitives
from eir.setup.output_setup import al_output_objects_as_dict
from eir.train_utils.distributed import (
    AttrDelegatedDistributedDataParallel,
    only_call_on_master_node,
)
from eir.train_utils.evaluation import validation_handler
from eir.train_utils.lr_scheduling import attach_lr_scheduler, set_up_lr_scheduler
from eir.train_utils.metrics import (
    MetricRecord,
    al_metric_record_dict,
    get_average_history_filepath,
    get_buffered_metrics_writer,
    get_metrics_dataframes,
    get_metrics_files,
    persist_metrics,
    read_metrics_history_file,
)
from eir.train_utils.optim import AttrDelegatedSWAWrapper
from eir.train_utils.utils import get_run_folder, validate_handler_dependencies
from eir.utils.logging import get_logger
from eir.visualization import visualization_funcs as vf

if TYPE_CHECKING:
    from eir.setup.config import Configs
    from eir.train import Experiment
    from eir.train_utils.metrics import al_step_metric_dict

# Aliases
al_handler = Callable[[Engine, "HandlerConfig"], None]
al_event = EventsList | CallableEventWithFilter
al_handler_and_event = Tuple[al_handler, al_event]
al_sample_interval_handlers = Tuple[al_handler_and_event, ...]


logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class HandlerConfig:
    experiment: "Experiment"
    run_folder: Path
    output_folder: str
    monitoring_metrics: List[Tuple[str, str, str]]


def configure_trainer(
    trainer: Engine,
    experiment: "Experiment",
    validation_handler_callable: al_handler = validation_handler,
) -> Engine:
    gc = experiment.configs.global_config
    run_folder = get_run_folder(output_folder=gc.be.output_folder)

    monitoring_metrics = _get_monitoring_metrics(
        outputs_as_dict=experiment.outputs, metric_record_dict=experiment.metrics
    )

    handler_config = HandlerConfig(
        experiment=experiment,
        run_folder=run_folder,
        output_folder=gc.be.output_folder,
        monitoring_metrics=monitoring_metrics,
    )

    train_monitoring_metrics = _parse_metrics_for_train_running_average(
        monitoring_metrics=monitoring_metrics
    )
    _call_and_undo_ignite_local_rank_side_effects(
        func=_attach_running_average_metrics,
        kwargs={"engine": trainer, "monitoring_metrics": train_monitoring_metrics},
    )

    _maybe_attach_progress_bar(trainer=trainer, do_not_attach=gc.vl.no_pbar)

    _attach_sample_interval_handlers(
        trainer=trainer,
        handler_config=handler_config,
        validation_handler_callable=validation_handler_callable,
    )

    if gc.tc.early_stopping_patience:
        _attach_early_stopping_handler(trainer=trainer, handler_config=handler_config)

    # TODO: Implement warmup for same LR scheduling
    if gc.lr_schedule != "same":
        lr_scheduler = set_up_lr_scheduler(handler_config=handler_config)
        attach_lr_scheduler(
            engine=trainer, lr_scheduler=lr_scheduler, experiment=experiment
        )
    elif gc.lr_schedule == "same" and gc.lr.warmup_steps:
        raise NotImplementedError("Warmup not yet implemented for 'same' LR schedule.")

    if handler_config.output_folder:
        _attach_run_event_handlers(trainer=trainer, handler_config=handler_config)

    return trainer


@only_call_on_master_node
def _attach_sample_interval_handlers(
    trainer: Engine,
    handler_config: "HandlerConfig",
    validation_handler_callable: Callable = validation_handler,
) -> Engine:
    exp = handler_config.experiment
    gc = exp.configs.global_config

    try:
        iter_per_epoch = len(exp.train_loader)
    except TypeError:
        iter_per_epoch = None

    validation_handler_and_event = _get_validation_handler_and_event(
        sample_interval_base=gc.ec.sample_interval,
        iter_per_epoch=iter_per_epoch,
        n_epochs=gc.be.n_epochs,
        early_stopping_patience=gc.tc.early_stopping_patience,
        validation_handler_callable=validation_handler_callable,
    )
    all_handler_events = [validation_handler_and_event]

    if gc.aa.compute_attributions:

        try:
            iter_per_epoch = len(exp.train_loader)
        except TypeError:
            iter_per_epoch = None

        attribution_handler_and_event = _get_attribution_handler_and_event(
            iter_per_epoch=iter_per_epoch,
            n_epochs=gc.be.n_epochs,
            sample_interval_base=gc.ec.sample_interval,
            attributions_every_sample_factor=gc.aa.attributions_every_sample_factor,
            early_stopping_patience=gc.tc.early_stopping_patience,
        )
        all_handler_events.append(attribution_handler_and_event)

    for handler, event in all_handler_events:
        trainer.add_event_handler(
            event_name=event,
            handler=handler,
            handler_config=handler_config,
        )

    return trainer


def _get_validation_handler_and_event(
    sample_interval_base: int,
    iter_per_epoch: Optional[int],
    n_epochs: int,
    early_stopping_patience: int,
    validation_handler_callable: Callable,
) -> al_handler_and_event:
    validation_event: al_event = Events.ITERATION_COMPLETED(every=sample_interval_base)

    if iter_per_epoch is None:
        return validation_handler_callable, validation_event

    do_run_when_training_complete = _do_run_completed_handler(
        iter_per_epoch=iter_per_epoch,
        n_epochs=n_epochs,
        sample_interval=sample_interval_base,
    )

    if do_run_when_training_complete and not early_stopping_patience:
        validation_event = validation_event | Events.COMPLETED

    return validation_handler_callable, validation_event


def _get_attribution_handler_and_event(
    iter_per_epoch: Optional[int],
    n_epochs: int,
    sample_interval_base: int,
    attributions_every_sample_factor: int,
    early_stopping_patience: int,
) -> al_handler_and_event:
    attribution_handler_callable = attribution_analysis_handler

    if attributions_every_sample_factor == 0:
        attribution_event: al_event = Events.COMPLETED
        logger.debug("Attributions will be computed at run end.")

        return attribution_handler_callable, attribution_event

    attribution_handler_interval = (
        sample_interval_base * attributions_every_sample_factor
    )
    attribution_event = Events.ITERATION_COMPLETED(every=attribution_handler_interval)

    if iter_per_epoch is None:
        return attribution_handler_callable, attribution_event

    do_run_when_training_complete = _do_run_completed_handler(
        iter_per_epoch=iter_per_epoch,
        n_epochs=n_epochs,
        sample_interval=attribution_handler_interval,
    )

    if do_run_when_training_complete and not early_stopping_patience:
        attribution_event = attribution_event | Events.COMPLETED

    logger.debug(
        "Attributions will be computed every %d iterations.",
        attribution_handler_interval,
    )

    return attribution_handler_callable, attribution_event


def _do_run_completed_handler(iter_per_epoch: int, n_epochs: int, sample_interval: int):
    """
    We need this function to avoid running handlers twice in cases where the total
    number of iterations has a remainder of 0 w.r.t. the sample interval.
    """
    if (iter_per_epoch * n_epochs) % sample_interval != 0:
        return True

    return False


def _get_monitoring_metrics(
    outputs_as_dict: al_output_objects_as_dict,
    metric_record_dict: al_metric_record_dict,
) -> List[Tuple[str, str, str]]:
    """
    The spec for the tuple here follows the metric dict spec, i.e. the tuple is:
    (column_name, metric).
    """
    target_columns_gen = get_output_info_generator(outputs_as_dict=outputs_as_dict)

    loss_average_metrics: tuple[str, str, str] = ("average", "average", "loss-average")
    perf_average_metrics: tuple[str, str, str] = ("average", "average", "perf-average")
    monitoring_metrics: list[tuple[str, str, str]] = [
        loss_average_metrics,
        perf_average_metrics,
    ]

    def _parse_target_metrics(
        output_name_: str, metric_name: str, column_name_: str
    ) -> str:
        return f"{output_name_}_{column_name_}_{metric_name}"

    for output_name, output_target_type, column_name in target_columns_gen:
        if output_target_type in ("con", "cat"):
            cur_output_object = outputs_as_dict[output_name]
            cur_output_type = cur_output_object.output_config.output_info.output_type
            assert cur_output_type == "tabular"

            al_record = tuple[MetricRecord, ...]
            cur_record = metric_record_dict[output_target_type]
            assert isinstance(cur_record, tuple)
            cur_metric_records: al_record = cur_record

            for metric in cur_metric_records:
                if not metric.only_val:
                    parsed_metric = _parse_target_metrics(
                        output_name_=output_name,
                        column_name_=column_name,
                        metric_name=metric.name,
                    )
                    cur_tuple: tuple[str, str, str] = (
                        output_name,
                        column_name,
                        parsed_metric,
                    )
                    monitoring_metrics.append(cur_tuple)

        # manually add loss record as it's not in metric records, but from criteria
        loss_name = _parse_target_metrics(
            output_name_=output_name, metric_name="loss", column_name_=column_name
        )
        metrics_keys = (output_name, column_name, loss_name)
        monitoring_metrics.append(metrics_keys)

    return monitoring_metrics


@validate_handler_dependencies(
    [validation_handler],
)
@only_call_on_master_node
def _attach_early_stopping_handler(trainer: Engine, handler_config: "HandlerConfig"):
    gc = handler_config.experiment.configs.global_config

    early_stopping_handler = _get_early_stopping_handler(
        trainer=trainer,
        handler_config=handler_config,
        patience_steps=gc.tc.early_stopping_patience,
    )

    early_stopping_event_kwargs = _get_early_stopping_event_filter_kwargs(
        early_stopping_iteration_buffer=gc.tc.early_stopping_buffer,
        sample_interval=gc.ec.sample_interval,
    )

    assert isinstance(early_stopping_event_kwargs, dict)
    trainer.add_event_handler(
        event_name=Events.ITERATION_COMPLETED(
            **early_stopping_event_kwargs,
        ),
        handler=early_stopping_handler,
    )

    return trainer


def _get_early_stopping_handler(
    trainer: Engine, handler_config: HandlerConfig, patience_steps: int
):
    scoring_function = _get_latest_validation_value_score_function(
        run_folder=handler_config.run_folder, column="perf-average"
    )

    logger.debug(
        "Setting early stopping patience to %d validation steps.", patience_steps
    )

    handler = EarlyStopping(
        patience=patience_steps, score_function=scoring_function, trainer=trainer
    )
    handler.logger = logger

    return handler


@overload
def _get_early_stopping_event_filter_kwargs(
    early_stopping_iteration_buffer: None, sample_interval: int
) -> dict[str, int]: ...


@overload
def _get_early_stopping_event_filter_kwargs(
    early_stopping_iteration_buffer: int, sample_interval: int
) -> dict[str, Callable[[Engine, int], bool]]: ...


def _get_early_stopping_event_filter_kwargs(
    early_stopping_iteration_buffer, sample_interval
):
    if early_stopping_iteration_buffer is None:
        return {"every": sample_interval}

    logger.info(
        "Early stopping checks will be activated after %d iterations.",
        early_stopping_iteration_buffer,
    )
    has_checked = False

    def _early_stopping_event_filter(engine: Engine, event: int) -> bool:
        iteration = event

        if iteration < early_stopping_iteration_buffer:
            return False

        nonlocal has_checked
        if not has_checked:
            logger.debug(
                "%d iterations done, early stopping checks activated from now on.",
                iteration,
            )
            has_checked = True

        if iteration % sample_interval == 0:
            return True

        return False

    return {"event_filter": _early_stopping_event_filter}


def _get_latest_validation_value_score_function(
    run_folder: Path, column: str
) -> Callable[[Engine], float]:
    eval_history_fpath = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix="validation_"
    )

    def scoring_function(engine: Engine) -> float:
        eval_df = read_metrics_history_file(eval_history_fpath)
        latest_val_loss = eval_df[column].iloc[-1]
        return latest_val_loss

    return scoring_function


def _parse_metrics_for_train_running_average(
    monitoring_metrics: List[Tuple[str, str, str]]
) -> List[Tuple[str, str, str]]:
    """
    We skip computing performance averages on training batches,
    as some metrics (e.g. ROC-AUC, R2) can be undefined (e.g. batch with only
    positive or negative labels) or fluctuate a lot (e.g. R2).
    """
    to_skip = ("perf-average",)

    parsed_metrics: List[Tuple[str, str, str]] = []

    for output_name, column_name, metric_name in monitoring_metrics:
        if metric_name in to_skip:
            continue

        parsed_metrics.append((output_name, column_name, metric_name))

    return parsed_metrics


class MyRunningAverage(RunningAverage):
    def __init__(
        self,
        src: Optional[Metric] = None,
        alpha: float = 0.98,
        output_transform: Optional[Callable] = None,
        epoch_bound: Optional[bool] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__(
            src=src,
            alpha=alpha,
            output_transform=output_transform,
            epoch_bound=epoch_bound,
            device=device,
        )

    def update(self, output: Union[torch.Tensor, float]) -> None:
        """
        We have this checking for nan to avoid corrupting / making the running
        average always become nan if a NaN output value (due to an empty output
        batch for a given target) being encountered.
        """
        if isinstance(output, torch.Tensor):
            if torch.isnan(output).any():
                return
        else:
            if np.isnan(output):
                return

        super().update(output)


def _attach_running_average_metrics(
    engine: Engine, monitoring_metrics: List[Tuple[str, str, str]]
) -> None:
    """
    For each metric, we create an output_transform function that grabs the
    target variable from the output of the step function (which is a dict).

    Basically what we attach to the trainer operates on the output of the
    update / step function, that we pass to the Engine definition.

    We use a partial so each lambda has its own metric variable (otherwise
    they all reference the same object as it gets overwritten).
    """
    for output_name, column_name, metric_name in monitoring_metrics:

        def output_transform(
            metric_dict_from_step: "al_step_metric_dict",
            output_name_key: str,
            column_name_key: str,
            metric_name_key: str,
        ) -> float:
            value = metric_dict_from_step[output_name_key][column_name_key][
                metric_name_key
            ]
            return value

        partial_func = partial(
            output_transform,
            output_name_key=output_name,
            column_name_key=column_name,
            metric_name_key=metric_name,
        )

        running_average = MyRunningAverage(
            output_transform=partial_func,
            alpha=0.98,
        )

        running_average.attach(
            engine,
            name=metric_name,
            usage=RunningBatchWise(),
        )


def _call_and_undo_ignite_local_rank_side_effects(func: Callable, kwargs: Dict):
    """
    This weird function is needed in the case where a GPU is available, calling some
    functions will trigger obscure ignite side effects that change some environment
    variables without warning.
    """
    original_local_rank = int(os.environ.get("LOCAL_RANK", 0))

    result = func(**kwargs)

    cur_local_rank = int(os.environ.get("LOCAL_RANK", 0))
    if cur_local_rank != original_local_rank:
        logger.debug(
            "Enforcing local rank to be '%d' after ignite side effects",
            original_local_rank,
        )
        os.environ["LOCAL_RANK"] = str(original_local_rank)

    return result


@only_call_on_master_node
def _maybe_attach_progress_bar(trainer: Engine, do_not_attach: bool) -> None:
    do_attach = not do_not_attach
    if do_attach:
        pbar = ProgressBar()
        pbar.attach(engine=trainer, metric_names=["loss-average"])
        trainer.add_event_handler(
            event_name=Events.EPOCH_COMPLETED,
            handler=_log_stats_to_pbar,
            pbar=pbar,
        )


def _log_stats_to_pbar(engine: Engine, pbar: ProgressBar) -> None:
    log_string = f"[Epoch {engine.state.epoch}/{engine.state.max_epochs}]"

    key = "loss-average"
    value = engine.state.metrics[key]
    log_string += f" | {key}: {value:.4g}"

    pbar.log_message(log_string)


@only_call_on_master_node
def _attach_run_event_handlers(trainer: Engine, handler_config: HandlerConfig):
    exp = handler_config.experiment
    gc = handler_config.experiment.configs.global_config

    _save_yaml_configs(run_folder=handler_config.run_folder, configs=exp.configs)

    if gc.ec.checkpoint_interval is not None:
        trainer = _add_checkpoint_handler_wrapper(
            trainer=trainer,
            run_folder=handler_config.run_folder,
            output_folder=Path(gc.be.output_folder),
            n_to_save=gc.ec.n_saved_models,
            checkpoint_interval=gc.ec.checkpoint_interval,
            sample_interval=gc.ec.sample_interval,
            model=exp.model,
        )

    metric_writing_funcs = _get_metric_writing_funcs(
        sample_interval=gc.ec.sample_interval,
        outputs_as_dict=exp.outputs,
        run_folder=handler_config.run_folder,
        detail_level=gc.ec.saved_result_detail_level,
    )
    trainer.add_event_handler(
        event_name=Events.ITERATION_COMPLETED,
        handler=_write_training_metrics_handler,
        handler_config=handler_config,
        writer_funcs=metric_writing_funcs,
    )

    for plot_event in _get_plot_events(sample_interval=gc.ec.sample_interval):

        if gc.ec.saved_result_detail_level <= 1:
            continue

        try:
            iter_per_epoch = len(exp.train_loader)
        except TypeError:
            iter_per_epoch = None

        if plot_event == Events.COMPLETED and iter_per_epoch is None:
            continue

        if plot_event == Events.COMPLETED and not _do_run_completed_handler(
            iter_per_epoch=len(exp.train_loader),
            n_epochs=gc.be.n_epochs,
            sample_interval=gc.ec.sample_interval,
        ):
            continue

        trainer = _attach_plot_progress_handler(
            trainer=trainer,
            plot_event=plot_event,
            handler_config=handler_config,
        )

    if exp.hooks.custom_handler_attachers is not None:
        custom_handlers = _get_custom_handlers(handler_config=handler_config)
        trainer = _attach_custom_handlers(
            trainer=trainer,
            handler_config=handler_config,
            custom_handlers=custom_handlers,
        )

    return trainer


def _save_yaml_configs(run_folder: Path, configs: "Configs"):
    for config_name, config_object in configs.__dict__.items():

        if config_name == "gc":
            continue

        cur_output_path = Path(run_folder / "configs" / config_name).with_suffix(
            ".yaml"
        )
        aislib.misc_utils.ensure_path_exists(path=cur_output_path)

        config_object_as_primitives = object_to_primitives(obj=config_object)

        with open(str(cur_output_path), "w") as yaml_file_handle:
            yaml.dump(data=config_object_as_primitives, stream=yaml_file_handle)


def _add_checkpoint_handler_wrapper(
    trainer: Engine,
    run_folder: Path,
    output_folder: Path,
    n_to_save: Union[int, None],
    checkpoint_interval: int,
    sample_interval: int,
    model: nn.Module,
) -> Engine:
    checkpoint_score_function, score_name = None, None
    if n_to_save is not None:
        logger.debug(
            "Setting up scoring function for checkpoint interval, "
            "keeping top %d models.",
            n_to_save,
        )

        if checkpoint_interval % sample_interval != 0:
            raise ValueError(
                f"The checkpoint interval ({checkpoint_interval}) is not "
                f"a multiplication of the sample interval "
                f"({sample_interval}) and top n saved models are being "
                f"requested. This means the scoring functions would give"
                f"a wrong result for the checkpoint (as a validation step"
                f"is not performed at the checkpoint interval)."
            )

        score_name = "perf-average"
        checkpoint_score_function = _get_latest_validation_value_score_function(
            run_folder=run_folder, column=score_name
        )

    assert n_to_save is not None
    assert checkpoint_score_function is not None
    assert score_name is not None
    checkpoint_handler = _get_checkpoint_handler(
        run_folder=run_folder,
        output_folder=output_folder,
        n_to_save=n_to_save,
        score_function=checkpoint_score_function,
        score_name=score_name,
    )

    trainer = _attach_checkpoint_handler(
        trainer=trainer,
        checkpoint_handler=checkpoint_handler,
        checkpoint_interval=checkpoint_interval,
        model=model,
    )

    return trainer


def _get_metric_writing_funcs(
    sample_interval: int,
    outputs_as_dict: al_output_objects_as_dict,
    run_folder: Path,
    detail_level: int,
) -> Dict[str, Dict[str, Callable]]:
    buffer_interval = sample_interval // 2

    target_generator = get_output_info_generator(outputs_as_dict=outputs_as_dict)

    metrics_files = get_metrics_files(
        target_generator=target_generator,
        run_folder=run_folder,
        train_or_val_target_prefix="train_",
        detail_level=detail_level,
    )

    writer_funcs: Dict[str, Dict[str, Callable]] = {}
    for output_name, target_name_file_dict in metrics_files.items():
        writer_funcs[output_name] = {}

        for target_name, target_file in target_name_file_dict.items():
            writer_funcs[output_name][target_name] = get_buffered_metrics_writer(
                buffer_interval=buffer_interval
            )

    return writer_funcs


def _get_checkpoint_handler(
    run_folder: Path,
    output_folder: Path,
    n_to_save: int,
    score_function: Optional[Callable] = None,
    score_name: Optional[str] = None,
) -> ModelCheckpoint:
    def _default_global_step_transform(engine: Engine, event_name: str) -> int:
        return engine.state.iteration

    checkpoint_handler = ModelCheckpoint(
        dirname=str(Path(run_folder, "saved_models")),
        filename_prefix=output_folder.name,
        create_dir=True,
        score_name=score_name,
        n_saved=n_to_save,
        score_function=score_function,
        global_step_transform=_default_global_step_transform,
    )

    return checkpoint_handler


@validate_handler_dependencies(
    [validation_handler],
)
def _attach_checkpoint_handler(
    trainer: Engine,
    checkpoint_handler: ModelCheckpoint,
    checkpoint_interval: int,
    model: Union[
        nn.Module, AttrDelegatedSWAWrapper, AttrDelegatedDistributedDataParallel
    ],
) -> Engine:
    match model:
        case AttrDelegatedSWAWrapper() | AttrDelegatedDistributedDataParallel():
            model_to_save = model.module
        case _:
            model_to_save = model

    trainer.add_event_handler(
        event_name=Events.ITERATION_COMPLETED(every=checkpoint_interval),
        handler=checkpoint_handler,
        to_save={"model": model_to_save},
    )

    return trainer


def _write_training_metrics_handler(
    engine: Engine,
    handler_config: HandlerConfig,
    writer_funcs: Union[Dict[str, Callable], None] = None,
):
    """
    Note that trainer.state.metrics contains the *running averages* we are interested
    in.

    The main "problem" here is that we lose the structure of the metrics dict that
    we get from `train_utils.metrics.calculate_batch_metrics`, so we have to
    filter all metrics for a given target column specifically, from the 1d array
    trainer.state.metrics gives us.
    """

    iteration = engine.state.iteration

    is_first_iteration = True if iteration == 1 else False

    engine_output = engine.state.output
    engine_metrics_dict = engine.state.metrics
    assert isinstance(engine_output, dict)
    assert isinstance(engine_metrics_dict, dict)

    running_average_metrics = _unflatten_engine_metrics_dict(
        step_base=engine_output,
        engine_metrics_dict=engine_metrics_dict,
    )
    persist_metrics(
        handler_config=handler_config,
        metrics_dict=running_average_metrics,
        iteration=iteration,
        write_header=is_first_iteration,
        prefixes={"metrics": "train_", "writer": "train"},
        writer_funcs=writer_funcs,
    )


def _unflatten_engine_metrics_dict(
    step_base: "al_step_metric_dict", engine_metrics_dict: Dict[str, float]
) -> "al_step_metric_dict":
    """
    We need this to streamline the 1D dictionary that comes from engine.state.metrics.
    """

    nested_dict: "al_step_metric_dict" = {}

    for output_name, output_metric_dict in step_base.items():
        nested_dict[output_name] = {}

        for target_name, target_metric_dict in output_metric_dict.items():
            nested_dict[output_name][target_name] = {}

            for target_metric_name in target_metric_dict.keys():
                eng_run_avg_value = engine_metrics_dict[target_metric_name]
                nested_dict[output_name][target_name][
                    target_metric_name
                ] = eng_run_avg_value

    return nested_dict


def _get_plot_events(
    sample_interval: int,
) -> Tuple[events.CallableEventWithFilter, events.CallableEventWithFilter]:
    plot_events = (
        Events.ITERATION_COMPLETED(every=sample_interval),
        Events.COMPLETED,
    )
    return plot_events


@validate_handler_dependencies([_write_training_metrics_handler])
def _attach_plot_progress_handler(
    trainer: Engine,
    plot_event: events.CallableEventWithFilter,
    handler_config: HandlerConfig,
):
    trainer.add_event_handler(
        event_name=plot_event,
        handler=_plot_progress_handler,
        handler_config=handler_config,
    )

    return trainer


def _iterdir_ignore_hidden(path: Path) -> Iterator[Path]:
    """
    Mostly to avoid hidden files like .DS_Store on Macs.
    """
    for child in path.iterdir():
        if child.name.startswith("."):
            continue
        yield child


def _plot_progress_handler(engine: Engine, handler_config: HandlerConfig) -> None:
    gc = handler_config.experiment.configs.global_config

    # if no val data is available yet
    if engine.state.iteration < gc.ec.sample_interval:
        return

    run_folder = get_run_folder(output_folder=gc.be.output_folder)

    if gc.ec.saved_result_detail_level >= 3:
        for output_dir in _iterdir_ignore_hidden(path=run_folder / "results"):
            for target_dir in _iterdir_ignore_hidden(path=output_dir):
                target_column = target_dir.name

                train_history_df, valid_history_df = get_metrics_dataframes(
                    results_dir=target_dir,
                    target_string=target_column,
                )

                vf.generate_all_training_curves(
                    training_history_df=train_history_df,
                    valid_history_df=valid_history_df,
                    output_folder=target_dir,
                    title_extra=target_column,
                    plot_skip_steps=gc.vl.plot_skip_steps,
                )

    train_avg_history_df, valid_avg_history_df = get_metrics_dataframes(
        results_dir=run_folder, target_string="average"
    )

    vf.generate_all_training_curves(
        training_history_df=train_avg_history_df,
        valid_history_df=valid_avg_history_df,
        output_folder=run_folder,
        title_extra="Multi Task Average",
        plot_skip_steps=gc.vl.plot_skip_steps,
    )


def _get_custom_handlers(handler_config: "HandlerConfig"):
    custom_handlers = handler_config.experiment.hooks.custom_handler_attachers

    return custom_handlers


def _attach_custom_handlers(
    trainer: Engine, handler_config: "HandlerConfig", custom_handlers
):
    if not custom_handlers:
        return trainer

    for custom_handler_attacher in custom_handlers:
        trainer = custom_handler_attacher(trainer, handler_config)

    return trainer



import torch
from einops import rearrange
from torch import nn

from eir.models.layers.norm_layers import RMSNorm


class LinearAttention(nn.Module):
    """
    From https://github.com/lucidrains
    """

    def __init__(
        self,
        embed_dim: int,
        heads: int = 4,
        dim_head: int = 32,
    ):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        hidden_dim = dim_head * heads

        self.to_qkv = nn.Conv2d(
            in_channels=embed_dim,
            out_channels=hidden_dim * 3,
            kernel_size=1,
            bias=False,
        )

        self.to_out = nn.Sequential(
            nn.Conv2d(in_channels=hidden_dim, out_channels=embed_dim, kernel_size=1),
            RMSNorm(dim=embed_dim, normalize_dim=1),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape

        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(
            lambda t: rearrange(t, "b (h c) x y -> b h c (x y)", h=self.heads), qkv
        )

        q = q.softmax(dim=-2)
        k = k.softmax(dim=-1)

        q = q * self.scale

        context = torch.einsum("b h d n, b h e n -> b h d e", k, v)

        out = torch.einsum("b h d e, b h d n -> b h e n", context, q)
        out = rearrange(out, "b h c (x y) -> b (h c) x y", h=self.heads, x=h, y=w)

        return self.to_out(out)



from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Dict, Optional, Protocol, Tuple, Union

import pandas as pd


class DataFrameFunctionProtocol(Protocol):
    def __call__(
        self,
        *args: Any,
        df: pd.DataFrame,
        column_name: Optional[str] = None,
        **kwargs: Any
    ) -> pd.DataFrame: ...


@dataclass
class ColumnOperation:
    """
    function: Function to apply on the label dataframe.
    """

    function: DataFrameFunctionProtocol
    function_args: Dict[str, Any]
    extra_columns_deps: Union[Tuple[str, ...], None] = ()
    only_apply_if_target: bool = False
    column_dtypes: Union[Dict[str, Any], None] = None


def streamline_df(df_func: DataFrameFunctionProtocol) -> Callable[[Any], pd.DataFrame]:
    @wraps(df_func)
    def wrapper(*args, df=None, column_name=None, **kwargs) -> pd.DataFrame:
        df = df_func(*args, df=df, column_name=column_name, **kwargs)

        return df

    return wrapper



from argparse import Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Sequence, Union

import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import ensure_path_exists
from sklearn.preprocessing import LabelEncoder, StandardScaler

from eir.data_load.data_utils import get_output_info_generator
from eir.data_load.label_setup import al_label_transformers_object
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.train_utils.evaluation import PerformancePlotConfig
from eir.visualization import visualization_funcs as vf

if TYPE_CHECKING:
    from eir.predict import PredictExperiment


def predict_tabular_wrapper_with_labels(
    predict_config: "PredictExperiment",
    all_predictions: dict[str, dict[str, torch.Tensor]],
    all_labels: dict[str, dict[str, torch.Tensor]],
    all_ids: dict[str, dict[str, list[str]]],
    predict_cl_args: Namespace,
) -> None:
    target_columns_gen = get_output_info_generator(
        outputs_as_dict=predict_config.outputs
    )

    for output_name, target_head_name, target_column_name in target_columns_gen:
        if target_head_name == "general":
            continue

        target_predictions = all_predictions[output_name][target_column_name]
        predictions = _parse_predictions(target_predictions=target_predictions)

        cur_output_object = predict_config.outputs[output_name]
        assert isinstance(cur_output_object, ComputedTabularOutputInfo)

        target_transformers = cur_output_object.target_transformers
        cur_target_transformer = target_transformers[target_column_name]
        classes = _get_target_class_names(
            transformer=cur_target_transformer, target_column=target_column_name
        )

        output_folder = Path(
            predict_cl_args.output_folder, output_name, target_column_name
        )
        ensure_path_exists(path=output_folder, is_folder=True)

        target_labels = None
        if all_labels:
            target_labels = all_labels[output_name][target_column_name].cpu().numpy()

        cur_ids = all_ids[output_name][target_column_name]

        df_merged_predictions = _merge_ids_predictions_and_labels(
            ids=cur_ids,
            predictions=predictions,
            labels=target_labels,
            prediction_classes=classes,
        )

        df_predictions = _add_inverse_transformed_columns_to_predictions(
            df=df_merged_predictions,
            target_column_name=target_column_name,
            target_column_type=target_head_name,
            transformer=cur_target_transformer,
            evaluation=predict_cl_args.evaluate,
        )

        _save_predictions(
            df_predictions=df_predictions,
            output_folder=output_folder,
        )

        if predict_cl_args.evaluate:
            cur_labels = all_labels[output_name][target_column_name].cpu().numpy()

            plot_config = PerformancePlotConfig(
                val_outputs=predictions,
                val_labels=cur_labels,
                val_ids=cur_ids,
                iteration=0,
                column_name=target_column_name,
                column_type=target_head_name,
                target_transformer=cur_target_transformer,
                output_folder=output_folder,
            )

            vf.gen_eval_graphs(plot_config=plot_config)


def _merge_ids_predictions_and_labels(
    ids: Sequence[str],
    predictions: np.ndarray,
    labels: Optional[np.ndarray],
    prediction_classes: Union[Sequence[str], None] = None,
    label_column_name: str = "True Label",
) -> pd.DataFrame:
    df = pd.DataFrame()

    df["ID"] = ids
    df = df.set_index("ID")

    if labels is not None:
        df[label_column_name] = labels

    if prediction_classes is None:
        prediction_classes = [f"Score Class {i}" for i in range(predictions.shape[1])]

    df[prediction_classes] = predictions

    return df


def predict_tabular_wrapper_no_labels(
    predict_config: "PredictExperiment",
    all_predictions: dict[str, dict[str, torch.Tensor]],
    all_ids: dict[str, dict[str, list[str]]],
    predict_cl_args: Namespace,
) -> None:
    target_columns_gen = get_output_info_generator(
        outputs_as_dict=predict_config.outputs
    )

    for output_name, target_head_name, target_column_name in target_columns_gen:
        if target_head_name == "general":
            continue

        target_predictions = all_predictions[output_name][target_column_name]
        predictions = _parse_predictions(target_predictions=target_predictions)

        cur_output_object = predict_config.outputs[output_name]
        assert isinstance(cur_output_object, ComputedTabularOutputInfo)

        target_transformers = cur_output_object.target_transformers
        cur_target_transformer = target_transformers[target_column_name]
        classes = _get_target_class_names(
            transformer=cur_target_transformer, target_column=target_column_name
        )

        output_folder = Path(
            predict_cl_args.output_folder, output_name, target_column_name
        )
        ensure_path_exists(path=output_folder, is_folder=True)

        cur_ids = all_ids[output_name][target_column_name]

        df_predictions = _merge_ids_and_predictions(
            ids=cur_ids,
            predictions=predictions,
            prediction_classes=classes,
        )

        df_predictions = _add_inverse_transformed_columns_to_predictions(
            df=df_predictions,
            target_column_name=target_column_name,
            target_column_type=target_head_name,
            transformer=cur_target_transformer,
            evaluation=predict_cl_args.evaluate,
        )

        _save_predictions(
            df_predictions=df_predictions,
            output_folder=output_folder,
        )


def _merge_ids_and_predictions(
    ids: Sequence[str],
    predictions: np.ndarray,
    prediction_classes: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    df = pd.DataFrame()

    df["ID"] = ids
    df = df.set_index("ID")

    if prediction_classes is None:
        prediction_classes = [f"Score Class {i}" for i in range(predictions.shape[1])]

    df[prediction_classes] = predictions

    return df


def _add_inverse_transformed_columns_to_predictions(
    df: pd.DataFrame,
    target_column_name: str,
    target_column_type: str,
    transformer: al_label_transformers_object,
    evaluation: bool,
) -> pd.DataFrame:
    df_copy = df.copy()

    assert target_column_type in ["con", "cat"], target_column_type

    if evaluation:
        df = _add_inverse_transformed_column(
            df=df_copy,
            column_name="True Label",
            transformer=transformer,
        )

    if target_column_type == "con":
        df = _add_inverse_transformed_column(
            df=df,
            column_name=target_column_name,
            transformer=transformer,
        )

    return df


def _add_inverse_transformed_column(
    df: pd.DataFrame,
    column_name: str,
    transformer: al_label_transformers_object,
) -> pd.DataFrame:
    df_copy = df.copy()

    tt_it = transformer.inverse_transform
    values = np.asarray(df[column_name].values)
    col_name = f"{column_name} Untransformed"

    if isinstance(transformer, LabelEncoder):
        df_copy.insert(loc=0, column=col_name, value=tt_it(values))
    elif isinstance(transformer, StandardScaler):
        values_parsed = tt_it(values.reshape(-1, 1)).flatten()
        df_copy.insert(loc=0, column=col_name, value=values_parsed)
    else:
        raise NotImplementedError(f"Transformer {transformer} not supported.")

    return df_copy


def _parse_predictions(target_predictions: torch.Tensor) -> np.ndarray:
    predictions = target_predictions.cpu().numpy()
    return predictions


def _get_target_class_names(
    transformer: al_label_transformers_object, target_column: str
):
    if isinstance(transformer, LabelEncoder):
        return transformer.classes_
    return [target_column]


def _save_predictions(df_predictions: pd.DataFrame, output_folder: Path) -> None:
    df_predictions.to_csv(path_or_buf=str(output_folder / "predictions.csv"))



import argparse
import types
from collections import Counter
from copy import copy
from dataclasses import dataclass, field, fields, is_dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    Sequence,
    Tuple,
    Type,
    Union,
)

import configargparse
import yaml

from eir.models.fusion.fusion_identity import IdentityConfig
from eir.models.fusion.fusion_mgmoe import MGMoEModelConfig
from eir.models.input.array.array_models import (
    ArrayModelConfig,
    LCLModelConfig,
    get_array_config_dataclass_mapping,
)
from eir.models.input.image.image_models import ImageModelConfig
from eir.models.input.omics.omics_models import (
    OmicsModelConfig,
    get_omics_config_dataclass_mapping,
)
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
    SequenceModelConfig,
)
from eir.models.input.tabular.tabular import (
    SimpleTabularModelConfig,
    TabularModelConfig,
)
from eir.models.layers.mlp_layers import ResidualMLPConfig
from eir.models.output.array.array_output_modules import (
    ArrayOutputModuleConfig,
    CNNUpscaleModelConfig,
    LCLOutputModelConfig,
)
from eir.models.output.sequence.sequence_output_modules import (
    SequenceOutputModuleConfig,
    TransformerSequenceOutputModuleConfig,
)
from eir.models.output.tabular.linear import LinearOutputModuleConfig
from eir.models.output.tabular.mlp_residual import ResidualMLPOutputModuleConfig
from eir.models.output.tabular.tabular_output_modules import TabularOutputModuleConfig
from eir.setup import schemas
from eir.setup.config_setup_modules.config_setup_utils import (
    get_yaml_iterator_with_injections,
)
from eir.setup.config_setup_modules.output_config_setup_sequence import (
    get_configs_object_with_seq_output_configs,
)
from eir.setup.config_validation import validate_train_configs
from eir.setup.schema_modules.latent_analysis_schemas import LatentSamplingConfig
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.setup.schemas import (
    AttributionAnalysisConfig,
    BasicExperimentConfig,
    EvaluationCheckpointConfig,
    GlobalConfig,
    GlobalModelConfig,
    LRScheduleConfig,
    OptimizationConfig,
    SupervisedMetricsConfig,
    TrainingControlConfig,
    VisualizationLoggingConfig,
)
from eir.setup.tensor_broker_setup import set_up_tensor_broker_config
from eir.train_utils.utils import configure_global_eir_logging
from eir.utils.logging import get_logger

al_input_types = Union[
    schemas.OmicsInputDataConfig,
    schemas.TabularInputDataConfig,
    schemas.SequenceInputDataConfig,
    schemas.ByteInputDataConfig,
]


al_output_types_schema_map = dict[
    str,
    Union[
        Type[TabularOutputTypeConfig],
        Type,
    ],
]

al_output_module_config_class_getter = (
    Callable[[str], Union[schemas.al_output_module_configs_classes, Any]],
)

al_output_model_config_classes = (
    Type[ResidualMLPOutputModuleConfig]
    | Type[LinearOutputModuleConfig]
    | Type[TransformerSequenceOutputModuleConfig]
    | Type[LCLModelConfig]
    | Type[CNNUpscaleModelConfig]
)

al_output_model_configs = (
    ResidualMLPOutputModuleConfig
    | LinearOutputModuleConfig
    | TransformerSequenceOutputModuleConfig
    | LCLModelConfig
    | CNNUpscaleModelConfig
)
al_output_model_init_map = dict[str, dict[str, al_output_model_config_classes]]

logger = get_logger(name=__name__)


def get_configs():
    main_cl_args, extra_cl_args = get_main_cl_args()

    output_folder, log_level = get_output_folder_and_log_level_from_cl_args(
        main_cl_args=main_cl_args, extra_cl_args=extra_cl_args
    )

    configure_global_eir_logging(output_folder=output_folder, log_level=log_level)

    configs = generate_aggregated_config(
        cl_args=main_cl_args,
        extra_cl_args_overload=extra_cl_args,
    )

    configs_with_seq_outputs = get_configs_object_with_seq_output_configs(
        configs=configs,
    )
    validate_train_configs(configs=configs)

    return configs_with_seq_outputs


def get_main_cl_args() -> Tuple[argparse.Namespace, List[str]]:
    parser_ = get_main_parser()
    cl_args, extra_cl_args = parser_.parse_known_args()

    return cl_args, extra_cl_args


def get_output_folder_and_log_level_from_cl_args(
    main_cl_args: argparse.Namespace,
    extra_cl_args: List[str],
) -> Tuple[str, str]:
    global_configs = main_cl_args.global_configs

    output_folder = None
    for config in global_configs:
        with open(config, "r") as f:
            config_dict = yaml.safe_load(f)

            output_folder = config_dict.get("basic_experiment", {}).get("output_folder")
            log_level = config_dict.get("visualization_logging", {}).get(
                "log_level", "info"
            )

            if output_folder and log_level:
                break

    for arg in extra_cl_args:
        if "output_folder=" in arg:
            output_folder = arg.split("=")[1]
            break

    if output_folder is None:
        raise ValueError("Output folder not found in global configs.")

    return output_folder, log_level


def get_main_parser(
    global_nargs: Literal["+", "*"] = "+", output_nargs: Literal["+", "*"] = "+"
) -> configargparse.ArgumentParser:
    parser_ = configargparse.ArgumentParser(
        config_file_parser_class=configargparse.YAMLConfigFileParser
    )

    global_required = True if global_nargs == "+" else False
    output_required = True if output_nargs == "+" else False

    parser_.add_argument(
        "--global_configs",
        nargs=global_nargs,
        type=str,
        required=global_required,
        help="Global .yaml configurations for the experiment.",
    )

    parser_.add_argument(
        "--input_configs",
        type=str,
        nargs="*",
        required=False,
        default=[],
        help="Input feature extraction .yaml configurations. "
        "Each configuration represents one input.",
    )

    parser_.add_argument(
        "--fusion_configs",
        type=str,
        nargs="*",
        required=False,
        default=[],
        help="Fusion .yaml configurations.",
    )

    parser_.add_argument(
        "--output_configs",
        type=str,
        nargs=output_nargs,
        required=output_required,
        help="Output .yaml configurations.",
    )

    return parser_


def _recursive_search(
    dict_: Mapping, target: Any, path=None
) -> Generator[Tuple[str, Any], None, None]:
    if not path:
        path = []

    for key, value in dict_.items():
        local_path = copy(path)
        local_path.append(key)
        if isinstance(value, Mapping):
            yield from _recursive_search(dict_=value, target=target, path=local_path)
        else:
            if value == target:
                yield local_path, value


@dataclass
class Configs:
    global_config: schemas.GlobalConfig
    input_configs: Sequence[schemas.InputConfig]
    fusion_config: schemas.FusionConfig
    output_configs: Sequence[schemas.OutputConfig]

    gc: schemas.GlobalConfig = field(init=False, repr=False)

    def __post_init__(self):
        self.gc = self.global_config


def generate_aggregated_config(
    cl_args: Union[argparse.Namespace, types.SimpleNamespace],
    extra_cl_args_overload: Union[List[str], None] = None,
    strict: bool = True,
) -> Configs:
    global_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.global_configs, extra_cl_args=extra_cl_args_overload
    )
    global_config = get_global_config(global_configs=global_config_iter)

    inputs_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.input_configs,
        extra_cl_args=extra_cl_args_overload,
    )
    input_configs = get_input_configs(input_configs=inputs_config_iter)

    fusion_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.fusion_configs,
        extra_cl_args=extra_cl_args_overload,
    )
    fusion_config = load_fusion_configs(fusion_configs=fusion_config_iter)

    output_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.output_configs,
        extra_cl_args=extra_cl_args_overload,
    )
    output_configs = load_output_configs(output_configs=output_config_iter)

    if strict:
        _check_input_and_output_config_names(
            input_configs=input_configs, output_configs=output_configs
        )

    aggregated_configs = Configs(
        global_config=global_config,
        input_configs=input_configs,
        fusion_config=fusion_config,
        output_configs=output_configs,
    )

    return aggregated_configs


def get_global_config(global_configs: Iterable[dict]) -> GlobalConfig:
    combined_config = combine_dicts(dicts=global_configs)

    basic_experiment_config = BasicExperimentConfig(
        **combined_config.get("basic_experiment", {})
    )
    model_config = GlobalModelConfig(**combined_config.get("model", {}))
    optimization_config = OptimizationConfig(**combined_config.get("optimization", {}))
    lr_schedule_config = LRScheduleConfig(**combined_config.get("lr_schedule", {}))
    training_control_config = TrainingControlConfig(
        **combined_config.get("training_control", {})
    )
    evaluation_checkpoint_config = EvaluationCheckpointConfig(
        **combined_config.get("evaluation_checkpoint", {})
    )
    attribution_analysis_config = AttributionAnalysisConfig(
        **combined_config.get("attribution_analysis", {})
    )
    metrics_config = SupervisedMetricsConfig(**combined_config.get("metrics", {}))
    visualization_logging_config = VisualizationLoggingConfig(
        **combined_config.get("visualization_logging", {})
    )

    latent_sampling = None
    if "latent_sampling" in combined_config and isinstance(
        combined_config["latent_sampling"], dict
    ):
        latent_sampling = LatentSamplingConfig(**combined_config["latent_sampling"])

    global_config = GlobalConfig(
        basic_experiment=basic_experiment_config,
        model=model_config,
        optimization=optimization_config,
        lr_schedule=lr_schedule_config,
        training_control=training_control_config,
        evaluation_checkpoint=evaluation_checkpoint_config,
        attribution_analysis=attribution_analysis_config,
        metrics=metrics_config,
        visualization_logging=visualization_logging_config,
        latent_sampling=latent_sampling,
    )

    return modify_global_config(global_config)


def modify_global_config(global_config: GlobalConfig) -> GlobalConfig:
    gc_copy = copy(global_config)

    if gc_copy.basic_experiment.valid_size > 1.0:
        gc_copy.basic_experiment.valid_size = int(gc_copy.basic_experiment.valid_size)

    return gc_copy


def _maybe_add_latent_sampling_to_combined_config(combined_config: dict) -> dict:
    if "latent_sampling" in combined_config and isinstance(
        combined_config["latent_sampling"], dict
    ):
        combined_config["latent_sampling"] = LatentSamplingConfig(
            **combined_config["latent_sampling"]
        )

    return combined_config


def get_input_configs(
    input_configs: Iterable[dict],
) -> Sequence[schemas.InputConfig]:
    config_objects = []

    for config_dict in input_configs:
        config_object = init_input_config(yaml_config_as_dict=config_dict)
        config_objects.append(config_object)

    _check_input_config_names(input_configs=config_objects)
    return config_objects


def _check_input_config_names(input_configs: Iterable[schemas.InputConfig]) -> None:
    names = []
    for config in input_configs:
        names.append(config.input_info.input_name)

    if len(set(names)) != len(names):
        counts = Counter(names)
        raise ValueError(
            f"Found duplicates of input names in input configs: {counts}. "
            f"Please make sure each input source has a unique input_name."
        )


def validate_keys_against_dataclass(
    input_dict: Dict[str, Any], dataclass_type: Type, name: str = ""
) -> None:
    if not is_dataclass(dataclass_type):
        raise TypeError(f"Provided type {dataclass_type.__name__} is not a dataclass")

    expected_keys = {field_.name for field_ in fields(dataclass_type)}

    actual_keys = set(input_dict.keys())

    unexpected_keys = actual_keys - expected_keys
    if unexpected_keys:
        message = (
            f"Unexpected keys found in configuration: '{', '.join(unexpected_keys)}'. "
            f"Expected keys of type '{dataclass_type.__name__}': "
            f"'{', '.join(expected_keys)}'."
        )

        if name:
            message = f"{name}: {message}"

        raise KeyError(message)


def init_input_config(yaml_config_as_dict: Dict[str, Any]) -> schemas.InputConfig:
    cfg = yaml_config_as_dict

    validate_keys_against_dataclass(
        input_dict=cfg,
        dataclass_type=schemas.InputConfig,
        name=cfg.get("input_info", {}).get("input_name", ""),
    )

    input_info_object = schemas.InputDataConfig(**cfg["input_info"])

    input_schema_map = get_inputs_schema_map()
    input_type_info_class = input_schema_map[input_info_object.input_type]
    input_type_info_kwargs = cfg.get("input_type_info", {})
    input_type_info_object = input_type_info_class(**input_type_info_kwargs)

    _validate_input_type_info_object(
        input_type_info_object=input_type_info_object,
        input_source=input_info_object.input_source,
    )

    model_config = set_up_input_feature_extractor_config(
        input_info_object=input_info_object,
        input_type_info_object=input_type_info_object,
        model_init_kwargs_base=cfg.get("model_config", {}),
    )

    pretrained_config = set_up_pretrained_config(
        pretrained_config_dict=cfg.get("pretrained_config", None)
    )

    interpretation_config = set_up_interpretation_config(
        input_type=input_info_object.input_type,
        interpretation_config_dict=cfg.get("interpretation_config", None),
    )

    tensor_broker_config = set_up_tensor_broker_config(
        tensor_broker_config=cfg.get("tensor_broker_config", {})
    )

    input_config = schemas.InputConfig(
        input_info=input_info_object,
        input_type_info=input_type_info_object,
        pretrained_config=pretrained_config,
        model_config=model_config,
        interpretation_config=interpretation_config,
        tensor_broker_config=tensor_broker_config,
    )

    return input_config


def get_inputs_schema_map() -> Dict[
    str,
    Union[
        Type[schemas.OmicsInputDataConfig],
        Type[schemas.TabularInputDataConfig],
        Type[schemas.SequenceInputDataConfig],
        Type[schemas.ByteInputDataConfig],
    ],
]:
    mapping = {
        "omics": schemas.OmicsInputDataConfig,
        "tabular": schemas.TabularInputDataConfig,
        "sequence": schemas.SequenceInputDataConfig,
        "bytes": schemas.ByteInputDataConfig,
        "image": schemas.ImageInputDataConfig,
        "array": schemas.ArrayInputDataConfig,
    }

    return mapping


def _validate_input_type_info_object(
    input_type_info_object: al_input_types, input_source: str
) -> None:
    ito = input_type_info_object
    match ito:
        case schemas.TabularInputDataConfig():
            con = ito.input_con_columns
            cat = ito.input_cat_columns
            common = set(con).intersection(set(cat))
            if common:
                raise ValueError(
                    f"Found columns passed in both continuous and categorical inputs: "
                    f"{common}. In input source: {input_source}."
                    f"Please make sure that each column is only in one of the two, "
                    f"or create differently named copies of the columns."
                )


def set_up_input_feature_extractor_config(
    input_info_object: schemas.InputDataConfig,
    input_type_info_object: al_input_types,
    model_init_kwargs_base: dict,
) -> schemas.al_feature_extractor_configs:
    input_type = input_info_object.input_type

    model_config_class = get_input_feature_extractor_config_class(input_type=input_type)

    model_type = model_init_kwargs_base.get("model_type", None)
    if not model_type:
        try:
            model_type = getattr(model_config_class, "model_type")
        except AttributeError:
            raise AttributeError(
                "Not model type specified in model config and could not find default "
                "value for '%s'.",
                input_type,
            )

        logger.info(
            "Input model type not specified in model configuration for input name "
            "'%s', attempting to grab default value.",
            input_info_object.input_name,
        )

    model_type_init_config = set_up_feature_extractor_init_config(
        input_info_object=input_info_object,
        input_type_info_object=input_type_info_object,
        model_init_kwargs=model_init_kwargs_base.get("model_init_config", {}),
        model_type=model_type,
    )

    common_kwargs = {
        "model_type": model_type,
        "model_init_config": model_type_init_config,
    }
    other_specific_kwargs = {
        k: v for k, v in model_init_kwargs_base.items() if k not in common_kwargs
    }
    model_config_kwargs = {**common_kwargs, **other_specific_kwargs}
    model_config = model_config_class(**model_config_kwargs)

    return model_config


def get_input_feature_extractor_config_class(
    input_type: str,
) -> schemas.al_feature_extractor_configs_classes:
    model_config_setup_map = get_input_feature_extractor_config_init_class_map()

    return model_config_setup_map[input_type]


def get_input_feature_extractor_config_init_class_map() -> (
    Dict[str, schemas.al_feature_extractor_configs_classes]
):
    mapping = {
        "tabular": TabularModelConfig,
        "omics": OmicsModelConfig,
        "sequence": SequenceModelConfig,
        "bytes": SequenceModelConfig,
        "image": ImageModelConfig,
        "array": ArrayModelConfig,
    }

    return mapping


def set_up_feature_extractor_init_config(
    input_info_object: schemas.InputDataConfig,
    input_type_info_object: al_input_types,
    model_init_kwargs: dict,
    model_type: str,
) -> Dict:
    if getattr(input_type_info_object, "pretrained_model", None):
        return {}

    not_from_eir = get_is_not_eir_model_condition(
        input_info_object=input_info_object,
        model_type=model_type,
    )
    if not_from_eir:
        return model_init_kwargs

    if not model_init_kwargs:
        model_init_kwargs = {}

    model_config_map = get_feature_extractor_config_type_init_callable_map()
    model_config_callable = model_config_map[model_type]

    model_config = model_config_callable(**model_init_kwargs)

    return model_config


def get_is_not_eir_model_condition(
    input_info_object: schemas.InputDataConfig, model_type: str
) -> bool:
    is_possibly_external = getattr(input_info_object, "input_type") in (
        "sequence",
        "bytes",
        "image",
    )
    is_unknown_model = model_type not in ("sequence-default", "cnn", "lcl")
    not_from_eir = is_possibly_external and is_unknown_model
    return not_from_eir


def get_feature_extractor_config_type_init_callable_map() -> Dict[str, Type]:
    omics_mapping = get_omics_config_dataclass_mapping()
    array_mapping = get_array_config_dataclass_mapping()
    other_mapping = {
        "tabular": SimpleTabularModelConfig,
        "sequence-default": BasicTransformerFeatureExtractorModelConfig,
    }

    mapping = omics_mapping | array_mapping | other_mapping
    return mapping


def set_up_pretrained_config(
    pretrained_config_dict: Union[None, Dict[str, Any]]
) -> Union[None, schemas.BasicPretrainedConfig]:
    if pretrained_config_dict is None:
        return None

    config_class = get_pretrained_config_class()
    if config_class is None:
        return None

    config_object = config_class(**pretrained_config_dict)

    return config_object


def get_pretrained_config_class() -> Type[schemas.BasicPretrainedConfig]:
    return schemas.BasicPretrainedConfig


def set_up_interpretation_config(
    input_type: str, interpretation_config_dict: Union[None, Dict[str, Any]]
) -> Union[None, schemas.BasicInterpretationConfig]:
    config_class = get_interpretation_config_class(input_type=input_type)
    if config_class is None:
        return None

    if interpretation_config_dict is None:
        interpretation_config_dict = {}

    config_object = config_class(**interpretation_config_dict)

    return config_object


def get_interpretation_config_class(
    input_type: str,
) -> Union[None, Type[schemas.BasicInterpretationConfig]]:
    mapping = get_interpretation_config_schema_map()

    return mapping.get(input_type, None)


def get_interpretation_config_schema_map() -> (
    Dict[str, Type[schemas.BasicInterpretationConfig]]
):
    mapping = {
        "sequence": schemas.BasicInterpretationConfig,
        "image": schemas.BasicInterpretationConfig,
    }

    return mapping


def load_fusion_configs(fusion_configs: Iterable[dict]) -> schemas.FusionConfig:
    combined_config = combine_dicts(dicts=fusion_configs)

    validate_keys_against_dataclass(
        input_dict=combined_config, dataclass_type=schemas.FusionConfig
    )

    combined_config.setdefault("model_type", "mlp-residual")
    combined_config.setdefault("model_config", {})

    fusion_model_type = combined_config["model_type"]

    model_dataclass_config_class: (
        Type[ResidualMLPConfig] | Type[MGMoEModelConfig] | Type[IdentityConfig]
    ) = ResidualMLPConfig
    if fusion_model_type == "mgmoe":
        model_dataclass_config_class = MGMoEModelConfig
    elif fusion_model_type == "linear":
        model_dataclass_config_class = IdentityConfig

    fusion_config_kwargs = combined_config["model_config"]
    fusion_model_config = model_dataclass_config_class(**fusion_config_kwargs)

    tensor_broker_config = set_up_tensor_broker_config(
        tensor_broker_config=combined_config.get("tensor_broker_config", {})
    )

    fusion_config = schemas.FusionConfig(
        model_type=combined_config["model_type"],
        model_config=fusion_model_config,
        tensor_broker_config=tensor_broker_config,
    )

    return fusion_config


def load_output_configs(
    output_configs: Iterable[dict],
) -> Sequence[schemas.OutputConfig]:
    output_config_objects = []

    for config_dict in output_configs:
        config_object = init_output_config(
            yaml_config_as_dict=config_dict,
        )
        output_config_objects.append(config_object)

    _check_output_config_names(output_configs=output_config_objects)
    return output_config_objects


def _check_output_config_names(output_configs: Iterable[schemas.OutputConfig]) -> None:
    names = []
    for config in output_configs:
        names.append(config.output_info.output_name)

    if len(set(names)) != len(names):
        counts = Counter(names)
        raise ValueError(
            f"Found duplicates of input names in input configs: {counts}. "
            f"Please make sure each input source has a unique input_name."
        )


def init_output_config(
    yaml_config_as_dict: Dict[str, Any],
) -> schemas.OutputConfig:
    cfg = yaml_config_as_dict

    validate_keys_against_dataclass(
        input_dict=cfg,
        dataclass_type=schemas.OutputConfig,
        name=cfg.get("output_info", {}).get("output_name", ""),
    )

    output_info_object = schemas.OutputInfoConfig(**cfg["output_info"])

    output_schema_map = get_outputs_types_schema_map()
    output_type_info_class = output_schema_map[output_info_object.output_type]

    output_type_info_class_init_kwargs = cfg.get("output_type_info", {})
    output_type_info_object = output_type_info_class(
        **output_type_info_class_init_kwargs
    )

    model_config = set_up_output_module_config(
        output_info_object=output_info_object,
        model_init_kwargs_base=cfg.get("model_config", {}),
    )

    sampling_config = _set_up_basic_sampling_config(
        output_type_config=output_type_info_object,
        sampling_config=cfg.get("sampling_config", {}),
    )

    tensor_broker_config = set_up_tensor_broker_config(
        tensor_broker_config=cfg.get("tensor_broker_config", {})
    )

    output_config = schemas.OutputConfig(
        output_info=output_info_object,
        output_type_info=output_type_info_object,
        model_config=model_config,
        sampling_config=sampling_config,
        tensor_broker_config=tensor_broker_config,
    )

    return output_config


def _set_up_basic_sampling_config(
    output_type_config: schemas.al_output_type_configs, sampling_config: dict
) -> dict | schemas.ArrayOutputSamplingConfig | schemas.ImageOutputSamplingConfig:
    """
    Note that the sequence sampling config currently has it's own logic
    in output_config_setup_sequence.py.
    """
    sampling_config_object: (
        dict | schemas.ArrayOutputSamplingConfig | schemas.ImageOutputSamplingConfig
    )
    match output_type_config:
        case schemas.ArrayOutputTypeConfig():
            sampling_config_object = schemas.ArrayOutputSamplingConfig(
                **sampling_config
            )
        case schemas.ImageOutputTypeConfig():
            sampling_config_object = schemas.ImageOutputSamplingConfig(
                **sampling_config
            )

        case schemas.TabularOutputTypeConfig() | schemas.SequenceOutputTypeConfig():
            sampling_config_object = sampling_config
        case _:
            raise ValueError(f"Unknown output type config '{output_type_config}'.")

    return sampling_config_object


def get_outputs_types_schema_map() -> Dict[
    str,
    Type[schemas.TabularOutputTypeConfig]
    | Type[schemas.SequenceOutputTypeConfig]
    | Type[schemas.ArrayOutputTypeConfig]
    | Type[schemas.ImageOutputTypeConfig],
]:
    mapping = {
        "tabular": schemas.TabularOutputTypeConfig,
        "sequence": schemas.SequenceOutputTypeConfig,
        "array": schemas.ArrayOutputTypeConfig,
        "image": schemas.ImageOutputTypeConfig,
    }

    return mapping


def get_output_module_config_class(
    output_type: str,
) -> schemas.al_output_module_configs_classes:
    model_config_setup_map = get_output_module_config_class_map()

    return model_config_setup_map[output_type]


def get_output_module_config_class_map() -> (
    Dict[str, schemas.al_output_module_configs_classes]
):
    mapping = {
        "tabular": TabularOutputModuleConfig,
        "sequence": SequenceOutputModuleConfig,
        "array": ArrayOutputModuleConfig,
        "image": ArrayOutputModuleConfig,
    }

    return mapping


def set_up_output_module_config(
    output_info_object: schemas.OutputInfoConfig,
    model_init_kwargs_base: dict,
) -> schemas.al_output_module_configs:
    output_type = output_info_object.output_type

    model_config_class = get_output_module_config_class(output_type=output_type)

    model_type = None
    if model_init_kwargs_base:
        model_type = model_init_kwargs_base.get("model_type", None)

    if not model_type:
        try:
            model_type = getattr(model_config_class, "model_type")
        except AttributeError:
            raise AttributeError(
                "Not model type specified in model config and could not find default "
                "value for '%s'.",
                output_type,
            )

        logger.info(
            "Output model type not specified in model configuration with name '%s', "
            "attempting to grab default value.",
            output_info_object.output_name,
        )

    output_module_init_class_map = get_output_config_type_init_callable_map()

    model_type_config = set_up_output_module_init_config(
        model_init_kwargs_base=model_init_kwargs_base.get("model_init_config", {}),
        output_type=output_type,
        model_type=model_type,
        output_module_init_class_map=output_module_init_class_map,
    )

    common_kwargs = {"model_type": model_type, "model_init_config": model_type_config}
    other_specific_kwargs = {
        k: v for k, v in model_init_kwargs_base.items() if k not in common_kwargs
    }
    model_config_kwargs = {**common_kwargs, **other_specific_kwargs}
    model_config = model_config_class(**model_config_kwargs)

    return model_config


def set_up_output_module_init_config(
    model_init_kwargs_base: Union[None, dict],
    output_type: Literal["tabular", "sequence", "array"],
    model_type: str,
    output_module_init_class_map: al_output_model_init_map,
) -> al_output_model_configs:
    if not model_init_kwargs_base:
        model_init_kwargs_base = {}

    model_init_kwargs = copy(model_init_kwargs_base)

    model_init_config_callable = output_module_init_class_map[output_type][model_type]

    model_init_config = model_init_config_callable(**model_init_kwargs)

    return model_init_config


def get_output_config_type_init_callable_map() -> al_output_model_init_map:
    mapping: al_output_model_init_map = {
        "tabular": {
            "mlp_residual": ResidualMLPOutputModuleConfig,
            "linear": LinearOutputModuleConfig,
        },
        "sequence": {
            "sequence": TransformerSequenceOutputModuleConfig,
        },
        "array": {
            "lcl": LCLOutputModelConfig,
            "cnn": CNNUpscaleModelConfig,
        },
        "image": {
            "lcl": LCLOutputModelConfig,
            "cnn": CNNUpscaleModelConfig,
        },
    }

    return mapping


@dataclass
class TabularTargets:
    con_targets: Dict[str, Sequence[str]]
    cat_targets: Dict[str, Sequence[str]]

    def __len__(self):
        all_con = sum(len(i) for i in self.con_targets.values())
        all_cat = sum(len(i) for i in self.cat_targets.values())
        return all_con + all_cat


def get_all_tabular_targets(
    output_configs: Iterable[schemas.OutputConfig],
) -> TabularTargets:
    con_targets = {}
    cat_targets = {}

    for output_config in output_configs:
        if output_config.output_info.output_type != "tabular":
            continue

        output_name = output_config.output_info.output_name
        output_type_info = output_config.output_type_info
        assert isinstance(output_type_info, TabularOutputTypeConfig)
        con_targets[output_name] = output_type_info.target_con_columns
        cat_targets[output_name] = output_type_info.target_cat_columns

    targets = TabularTargets(con_targets=con_targets, cat_targets=cat_targets)
    return targets


def _check_input_and_output_config_names(
    input_configs: Sequence[schemas.InputConfig],
    output_configs: Sequence[schemas.OutputConfig],
    skip_keys: Sequence[str] = ("sequence", "array", "image"),
) -> None:
    """
    We allow for the same name to be used for sequence inputs and outputs,
    as it's used for specifying e.g. different model settings for the input feature
    extractor vs. the output module.
    """
    input_names = set(i.input_info.input_name for i in input_configs)
    output_names = set(
        i.output_info.output_name
        for i in output_configs
        if i.output_info.output_type not in skip_keys
    )

    intersection = output_names.intersection(input_names)
    if len(intersection) > 0:
        raise ValueError(
            "Found common names in input and output configs. Please ensure"
            " that there are no common names between the input and output configs."
            " Input config names: '%s'.\n"
            " Output config names: '%s'.\n",
            " Common names: '%s'.\n",
            input_names,
            output_names,
            intersection,
        )


def combine_dicts(dicts: Iterable[dict]) -> dict:
    combined_dict: dict = {}

    for dict_ in dicts:
        combined_dict = {**combined_dict, **dict_}

    return combined_dict



from dataclasses import dataclass
from typing import Literal, Union

from eir.models.output.tabular.linear import LinearOutputModuleConfig
from eir.models.output.tabular.mlp_residual import ResidualMLPOutputModuleConfig


@dataclass
class TabularOutputModuleConfig:
    """
    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param model_type:
         Which type of image model to use."""

    model_init_config: Union[ResidualMLPOutputModuleConfig, LinearOutputModuleConfig]
    model_type: Literal["mlp_residual", "linear"] = "mlp_residual"



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Literal, Protocol, Type, Union

from eir.models.input.array.models_cnn import CNNModel, CNNModelConfig
from eir.models.input.array.models_identity import IdentityModel, IdentityModelConfig
from eir.models.input.array.models_linear import LinearModel, LinearModelConfig
from eir.models.input.array.models_locally_connected import (
    FlattenFunc,
    LCLModel,
    LCLModelConfig,
    SimpleLCLModel,
    SimpleLCLModelConfig,
    flatten_h_w_fortran,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions

logger = get_logger(name=__name__)

al_omics_model_classes = Union[
    Type["CNNModel"],
    Type["LinearModel"],
    Type["SimpleLCLModel"],
    Type["LCLModel"],
    Type["IdentityModel"],
]

al_omics_models = Union[
    "CNNModel", "LinearModel", "SimpleLCLModel", "LCLModel", "IdentityModel"
]

al_omics_model_types = Literal[
    "cnn",
    "linear",
    "lcl-simple",
    "genome-local-net",
    "linear",
]

al_omics_model_config_classes = Union[
    Type[CNNModelConfig],
    Type[LinearModelConfig],
    Type[SimpleLCLModelConfig],
    Type[LCLModelConfig],
    Type[IdentityModelConfig],
]

al_omics_model_configs = Union[
    CNNModelConfig,
    LinearModelConfig,
    SimpleLCLModelConfig,
    LCLModelConfig,
    IdentityModelConfig,
]


@dataclass
class OmicsModelConfig:
    """
    :param model_type:
         Which type of image model to use.

    :param model_init_config:
          Configuration used to initialise model.
    """

    model_type: al_omics_model_types
    model_init_config: al_omics_model_configs


def get_omics_model_mapping() -> Dict[str, al_omics_model_classes]:
    mapping = {
        "cnn": CNNModel,
        "linear": LinearModel,
        "lcl-simple": SimpleLCLModel,
        "genome-local-net": LCLModel,
        "identity": IdentityModel,
    }

    return mapping


def get_omics_model_class(model_type: al_omics_model_types) -> al_omics_model_classes:
    mapping = get_omics_model_mapping()
    return mapping[model_type]


class Dataclass(Protocol):
    __dataclass_fields__: ClassVar[Dict]


def get_omics_config_dataclass_mapping() -> Dict[str, al_omics_model_config_classes]:
    mapping = {
        "cnn": CNNModelConfig,
        "linear": LinearModelConfig,
        "lcl-simple": SimpleLCLModelConfig,
        "genome-local-net": LCLModelConfig,
        "identity": IdentityModelConfig,
    }

    return mapping


def get_model_config_dataclass(
    model_type: str,
) -> al_omics_model_config_classes:
    mapping = get_omics_config_dataclass_mapping()
    return mapping[model_type]


def get_omics_model_init_kwargs(
    model_type: al_omics_model_types,
    model_config: al_omics_model_configs,
    data_dimensions: "DataDimensions",
) -> dict[str, Union["DataDimensions", al_omics_model_configs | FlattenFunc]]:
    """
    See: https://github.com/python/mypy/issues/5374 for type hint issue.

    Possibly split / extend this function later to account for other kwargs that just
    model_config, to allow for more flexibility in model instantiation (not restricting
    to just model_config object).
    """

    kwargs: dict[str, Union["DataDimensions", al_omics_model_configs | FlattenFunc]] = (
        {}
    )
    base_kwargs = model_config.__dict__
    base_kwargs = _enforce_omics_specific_settings(
        base_kwargs=base_kwargs, model_type=model_type
    )

    model_config_dataclass = get_model_config_dataclass(model_type=model_type)
    dataclass_instance = model_config_dataclass(**base_kwargs)

    kwargs["model_config"] = dataclass_instance
    kwargs["data_dimensions"] = data_dimensions

    match model_type:
        case "genome-local-net" | "lcl-simple":
            kwargs["flatten_fn"] = flatten_h_w_fortran

    return kwargs


def _enforce_omics_specific_settings(
    base_kwargs: Dict[str, Any], model_type: str
) -> Dict[str, Any]:
    expected = {
        "cnn": {
            "kernel_height": 1,
            "first_kernel_expansion_height": 4,
        }
    }

    if model_type not in expected:
        return base_kwargs

    for key, value in expected[model_type].items():
        logger.info(
            f"Overriding {key} to {value} for {model_type} in omics. "
            "If you want more control of these parameters, "
            "it might be a good idea to use the array input functionality."
        )
        base_kwargs[key] = value

    return base_kwargs



from dataclasses import dataclass
from typing import Dict

from sklearn.preprocessing import StandardScaler

from eir.data_load.label_setup import (
    al_label_transformers,
    al_target_columns,
    merge_target_columns,
)
from eir.setup import schemas
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)

al_num_outputs_per_target = Dict[str, int]


@dataclass
class ComputedTabularOutputInfo:
    output_config: schemas.OutputConfig
    num_outputs_per_target: al_num_outputs_per_target
    target_columns: al_target_columns
    target_transformers: al_label_transformers


def set_up_tabular_output(
    output_config: schemas.OutputConfig,
    target_transformers: Dict[str, al_label_transformers],
    *args,
    **kwargs,
) -> ComputedTabularOutputInfo:
    cur_target_transformers = target_transformers[output_config.output_info.output_name]
    num_outputs_per_target = set_up_num_outputs_per_target(
        target_transformers=cur_target_transformers
    )

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)

    target_columns = merge_target_columns(
        target_con_columns=list(output_type_info.target_con_columns),
        target_cat_columns=list(output_type_info.target_cat_columns),
    )

    tabular_output_info = ComputedTabularOutputInfo(
        output_config=output_config,
        num_outputs_per_target=num_outputs_per_target,
        target_columns=target_columns,
        target_transformers=cur_target_transformers,
    )

    return tabular_output_info


def set_up_num_outputs_per_target(
    target_transformers: al_label_transformers,
) -> al_num_outputs_per_target:
    num_outputs_per_target_dict = {}
    for target_column, transformer in target_transformers.items():
        if isinstance(transformer, StandardScaler):
            num_outputs = 1
        else:
            num_outputs = len(transformer.classes_)

            if num_outputs < 2:
                logger.warning(
                    f"Only {num_outputs} unique values found in categorical label "
                    f"column {target_column} (returned by {transformer}). This means "
                    f"that most likely an error will be raised if e.g. using "
                    f"nn.CrossEntropyLoss as it expects an output dimension of >=2."
                )

        num_outputs_per_target_dict[target_column] = num_outputs

    return num_outputs_per_target_dict



import math
import os
import reprlib
from dataclasses import dataclass
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from eir.data_load.data_source_modules.deeplake_ops import (
    is_deeplake_dataset,
    load_deeplake_dataset,
)
from eir.data_load.label_setup import (
    Labels,
    TabularFileInfo,
    al_all_column_ops,
    al_label_dict,
    al_label_transformers,
    al_target_label_dict,
    gather_ids_from_data_source,
    gather_ids_from_tabular_file,
    get_file_path_iterator,
    save_transformer_set,
    set_up_train_and_valid_tabular_data,
)
from eir.setup import schemas
from eir.setup.schema_modules.output_schemas_array import ArrayOutputTypeConfig
from eir.setup.schema_modules.output_schemas_sequence import SequenceOutputTypeConfig
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.train import Hooks


logger = get_logger(name=__name__)


def set_up_all_targets_wrapper(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    run_folder: Path,
    output_configs: Sequence[schemas.OutputConfig],
    hooks: Optional["Hooks"],
) -> "MergedTargetLabels":
    logger.info("Setting up target labels.")

    custom_ops = hooks.custom_column_label_parsing_ops if hooks else None
    target_labels = set_up_all_target_labels_wrapper(
        output_configs=output_configs,
        custom_label_ops=custom_ops,
        train_ids=train_ids,
        valid_ids=valid_ids,
    )
    save_transformer_set(
        transformers_per_source=target_labels.label_transformers,
        run_folder=run_folder,
    )

    return target_labels


@dataclass
class MissingTargetsInfo:
    missing_ids_per_modality: dict[str, set[str]]
    missing_ids_within_modality: dict[str, dict[str, set[str]]]
    precomputed_missing_ids: dict[str, dict[str, set[str]]]


def get_missing_targets_info(
    missing_ids_per_modality: dict[str, set[str]],
    missing_ids_within_modality: dict[str, dict[str, set[str]]],
    output_and_target_names: dict[str, list[str]],
) -> MissingTargetsInfo:

    precomputed_missing_ids = _precompute_missing_ids(
        missing_ids_per_modality=missing_ids_per_modality,
        missing_ids_within_modality=missing_ids_within_modality,
        output_and_target_names=output_and_target_names,
    )

    return MissingTargetsInfo(
        missing_ids_per_modality=missing_ids_per_modality,
        missing_ids_within_modality=missing_ids_within_modality,
        precomputed_missing_ids=precomputed_missing_ids,
    )


def _precompute_missing_ids(
    missing_ids_per_modality: dict[str, set[str]],
    missing_ids_within_modality: dict[str, dict[str, set[str]]],
    output_and_target_names: dict[str, list[str]],
) -> dict[str, dict[str, set[str]]]:
    """
    One could potentially optimize this space-wise by tracking modality-inner_key
    combinations that lead to the same set of missing IDs, then having them point
    to the same object (set of missing IDs) in memory.
    """

    precomputed_missing_ids: dict[str, dict[str, set[str]]] = {}

    for modality, target_names in output_and_target_names.items():
        if modality not in precomputed_missing_ids:
            precomputed_missing_ids[modality] = {}

        ids_this_modality = missing_ids_per_modality.get(modality, set())
        missing_ids_within = missing_ids_within_modality.get(modality, {})

        for target_name in target_names:
            cur_missing_within = missing_ids_within.get(target_name, set())
            combined_ids = ids_this_modality.union(cur_missing_within)

            if combined_ids:
                precomputed_missing_ids[modality][target_name] = combined_ids

    return precomputed_missing_ids


def get_all_output_and_target_names(
    output_configs: Sequence[schemas.OutputConfig],
) -> dict[str, list[str]]:
    output_and_target_names = {}

    for output_config in output_configs:
        output_name = output_config.output_info.output_name
        match output_config.output_type_info:
            case TabularOutputTypeConfig(
                target_con_columns=con_columns, target_cat_columns=cat_columns
            ):
                all_columns = list(con_columns) + list(cat_columns)
                output_and_target_names[output_name] = all_columns
            case ArrayOutputTypeConfig() | SequenceOutputTypeConfig():
                output_and_target_names[output_name] = [output_name]

    return output_and_target_names


def log_missing_targets_info(
    missing_targets_info: MissingTargetsInfo, all_ids: set[str]
) -> None:
    repr_formatter = reprlib.Repr()
    repr_formatter.maxset = 10

    logger.info(
        "Checking for missing target information. "
        "These will be ignored during loss and metric computation."
    )

    total_ids_count = len(all_ids)
    missing_within = missing_targets_info.missing_ids_within_modality
    max_columns_to_log = 5

    for modality, missing_ids in missing_targets_info.missing_ids_per_modality.items():
        missing_count = len(missing_ids)

        if missing_count == 0:
            logger.info(f"Output modality '{modality}' has no missing target IDs.")
            continue

        formatted_missing_ids = repr_formatter.repr(missing_ids)
        complete_count = total_ids_count - missing_count
        fraction_complete = (complete_count / total_ids_count) * 100

        logger.info(
            f"Missing target IDs for modality: '{modality}'\n"
            f"  - Missing IDs: {formatted_missing_ids}\n"
            f"  - Stats: Missing: {missing_count}, "
            f"Complete: {complete_count}/{total_ids_count} "
            f"({fraction_complete:.2f}% complete)\n"
        )

        if modality in missing_within:
            columns_logged = 0
            for target_column, ids in missing_within[modality].items():
                missing_within_count = len(ids)

                if missing_within_count == 0:
                    logger.info(
                        f"  - No missing target IDs in modality '{modality}', "
                        f"Column: '{target_column}'."
                    )
                    columns_logged += 1
                    if columns_logged >= max_columns_to_log:
                        break
                    continue

                if columns_logged >= max_columns_to_log:
                    additional_columns = (
                        len(missing_within[modality]) - max_columns_to_log
                    )
                    logger.info(
                        f"  - There are {additional_columns} "
                        f"more columns with missing IDs in modality '{modality}' "
                        f"not displayed."
                    )
                    break

                complete_within_count = total_ids_count - missing_within_count
                fraction_complete_within = (
                    complete_within_count / total_ids_count
                ) * 100

                formatted_ids = repr_formatter.repr(ids)
                logger.info(
                    f"  - Missing target IDs within modality '{modality}', "
                    f"Column: '{target_column}'\n"
                    f"      - Missing IDs: {formatted_ids}\n"
                    f"      - Stats: Missing: {missing_within_count}, "
                    f"Complete: {complete_within_count}/{total_ids_count} "
                    f"({fraction_complete_within:.2f}% complete)\n"
                )
                columns_logged += 1


@dataclass
class MergedTargetLabels:
    train_labels: al_target_label_dict
    valid_labels: al_target_label_dict
    label_transformers: dict[str, al_label_transformers]
    missing_ids_per_output: MissingTargetsInfo

    @property
    def all_labels(self):
        return {**self.train_labels, **self.valid_labels}


def set_up_all_target_labels_wrapper(
    output_configs: Sequence[schemas.OutputConfig],
    custom_label_ops: al_all_column_ops,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
) -> MergedTargetLabels:
    """
    We have different ways of setting up the missing labels information depending
    on the output type.

    For tabular, we have a csv file with the labels, so we can just read that in.
    Missing rows in the csv file are not included, and we can directly infer
    the missing IDs from full set of IDs compared to the IDs in the csv file.

    For sequence, the labels are computed on the fly during training, so we
    simply set up a dictionary with the IDs as keys and torch.nan as values.
    TODO:   Fill out the values here with the actual labels depending on available
            files / rows in sequence input .csv.

    For array, we read what files are available on the disk, and flag the missing
    ones as torch.nan.

    TODO:   There is some space inefficiency here, as for many output modalities,
            when we concatenate the dataframes, we will have NaNs in the columns
            that are not present in all output modalities.

    Note:   We store the dtypes here as e.g. from int columns, concatenating will
            create NaNs, which will convert the int columns to float.
    """
    df_labels_train = pd.DataFrame()
    df_labels_valid = pd.DataFrame()
    label_transformers = {}

    all_ids: set[str] = set(train_ids).union(set(valid_ids))
    per_modality_missing_ids: dict[str, set[str]] = {}
    within_modality_missing_ids: dict[str, dict[str, set[str]]] = {}

    dtypes: dict[str, dict[str, Any]] = {}

    tabular_target_labels_info = get_tabular_target_file_infos(
        output_configs=output_configs
    )

    for output_config in output_configs:
        output_source = output_config.output_info.output_source
        output_name = output_config.output_info.output_name
        output_type = output_config.output_info.output_type

        match output_type:
            case "tabular":
                tabular_info = tabular_target_labels_info[output_name]
                cur_labels = set_up_train_and_valid_tabular_data(
                    tabular_file_info=tabular_info,
                    custom_label_ops=custom_label_ops,
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                )
                cur_transformers = cur_labels.label_transformers
                label_transformers[output_name] = cur_transformers

                all_labels = cur_labels.all_labels
                cur_ids = set(all_labels.index)
                missing_ids = all_ids.difference(cur_ids)
                per_modality_missing_ids[output_name] = missing_ids

                logger.debug(
                    "Estimating missing IDs for tabular output %s.", output_name
                )
                missing_ids_per_target_column = compute_missing_ids_per_tabular_output(
                    all_labels_df=all_labels,
                    tabular_info=tabular_info,
                    output_name=output_name,
                )

                within_modality_missing_ids = {
                    **within_modality_missing_ids,
                    **missing_ids_per_target_column,
                }

            case "sequence":
                cur_labels = set_up_delayed_target_labels(
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    output_name=output_name,
                )

                logger.debug(
                    "Estimating missing IDs for sequence output %s.", output_name
                )
                missing_sequence_ids = find_sequence_output_missing_ids(
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    output_source=output_config.output_info.output_source,
                )

                per_modality_missing_ids[output_name] = missing_sequence_ids

            case "array" | "image":
                cur_labels = set_up_file_target_labels(
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    output_config=output_config,
                )

                logger.debug("Estimating missing IDs for array output %s.", output_name)
                cur_missing_ids = gather_torch_nan_missing_ids(
                    labels=cur_labels.all_labels,
                    output_name=output_name,
                )
                per_modality_missing_ids[output_name] = cur_missing_ids

                # this is needed as having missing modalities in deeplake
                # will cause conversion of int64 deeplake pointers to float64
                is_deeplake = is_deeplake_dataset(data_source=output_source)
                if is_deeplake:
                    dtypes[output_name] = {output_name: np.dtype("int64")}
                else:
                    dtypes[output_name] = {output_name: np.dtype("O")}

            case _:
                raise ValueError(f"Unknown output type: '{output_type}'.")

        df_train_cur = cur_labels.train_labels
        df_valid_cur = cur_labels.valid_labels

        df_train_cur["Output Name"] = output_name
        df_valid_cur["Output Name"] = output_name

        if output_name not in dtypes:
            dtypes[output_name] = df_train_cur.dtypes.to_dict()

        df_labels_train = pd.concat((df_labels_train, df_train_cur))
        df_labels_valid = pd.concat((df_labels_valid, df_valid_cur))

    df_labels_train = df_labels_train.set_index("Output Name", append=True)
    df_labels_valid = df_labels_valid.set_index("Output Name", append=True)

    df_labels_train = df_labels_train.dropna(how="all")
    df_labels_valid = df_labels_valid.dropna(how="all")

    dtypes = convert_dtypes(dtypes=dtypes)
    logger.debug("Converting train DF to dict.")
    train_labels_dict = df_to_nested_dict(df=df_labels_train, dtypes=dtypes)
    logger.debug("Converting valid DF to dict.")
    valid_labels_dict = df_to_nested_dict(df=df_labels_valid, dtypes=dtypes)

    output_and_target_names = get_all_output_and_target_names(
        output_configs=output_configs
    )
    missing_target_info = get_missing_targets_info(
        missing_ids_per_modality=per_modality_missing_ids,
        missing_ids_within_modality=within_modality_missing_ids,
        output_and_target_names=output_and_target_names,
    )
    log_missing_targets_info(missing_targets_info=missing_target_info, all_ids=all_ids)

    labels_data_object = MergedTargetLabels(
        train_labels=train_labels_dict,
        valid_labels=valid_labels_dict,
        label_transformers=label_transformers,
        missing_ids_per_output=missing_target_info,
    )

    return labels_data_object


def compute_missing_ids_per_tabular_output(
    all_labels_df: pd.DataFrame,
    tabular_info: TabularFileInfo,
    output_name: str = "output",
) -> Dict[str, Dict[str, set[str]]]:
    missing_per_target_column: Dict[str, Dict[str, set[str]]] = {output_name: {}}
    all_columns = list(tabular_info.con_columns) + list(tabular_info.cat_columns)

    for target_column in all_columns:
        cur_missing = set(all_labels_df.index[all_labels_df[target_column].isna()])
        missing_per_target_column[output_name][target_column] = cur_missing

    return missing_per_target_column


def set_up_delayed_target_labels(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_name: str,
) -> Labels:
    train_ids_set = set(train_ids)
    valid_ids_set = set(valid_ids)
    train_labels: al_label_dict = {
        id_: {output_name: torch.nan} for id_ in train_ids_set
    }
    valid_labels: al_label_dict = {
        id_: {output_name: torch.nan} for id_ in valid_ids_set
    }

    df_train = pd.DataFrame.from_dict(train_labels, orient="index")
    df_train["Output Name"] = output_name
    df_valid = pd.DataFrame.from_dict(valid_labels, orient="index")
    df_valid["Output Name"] = output_name

    return Labels(
        train_labels=df_train,
        valid_labels=df_valid,
        label_transformers={},
    )


def find_sequence_output_missing_ids(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_source: str,
) -> set[str]:

    seq_ids = set(gather_ids_from_data_source(data_source=Path(output_source)))

    train_ids_set = set(train_ids)
    valid_ids_set = set(valid_ids)
    all_ids = train_ids_set.union(valid_ids_set)

    missing_ids = all_ids.difference(seq_ids)

    return missing_ids


def set_up_file_target_labels(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_config: schemas.OutputConfig,
) -> Labels:
    """
    Note we have the .get(id_, torch.nan) here because we want to be able to
    handle if we have partially missing output modalities, e.g. in the case
    of output arrays, but we don't want to just drop those IDs.
    """
    output_name = output_config.output_info.output_name
    output_source = output_config.output_info.output_source
    output_name_inner_key = output_config.output_info.output_inner_key

    id_to_data_pointer_mapping = gather_data_pointers_from_data_source(
        data_source=Path(output_source),
        validate=True,
        output_inner_key=output_name_inner_key,
    )

    train_ids_set = set(train_ids)
    train_labels: al_label_dict = {
        id_: {output_name: id_to_data_pointer_mapping.get(id_, torch.nan)}
        for id_ in train_ids_set
    }

    valid_ids_set = set(valid_ids)
    valid_labels: al_label_dict = {
        id_: {output_name: id_to_data_pointer_mapping.get(id_, torch.nan)}
        for id_ in valid_ids_set
    }

    df_train = pd.DataFrame.from_dict(train_labels, orient="index")
    df_train["Output Name"] = output_name
    df_valid = pd.DataFrame.from_dict(valid_labels, orient="index")
    df_valid["Output Name"] = output_name

    return Labels(
        train_labels=df_train,
        valid_labels=df_valid,
        label_transformers={},
    )


def gather_torch_nan_missing_ids(labels: pd.DataFrame, output_name: str) -> set[str]:
    missing_ids = set()
    for id_, label in labels.iterrows():
        cur_label = label[output_name]
        if isinstance(cur_label, float) and math.isnan(cur_label):
            missing_ids.add(str(id_))

    return missing_ids


def convert_dtypes(dtypes: dict[str, dict[str, Any]]) -> dict[str, dict[str, Any]]:
    dtype_mapping = {
        "int64": int,
    }

    converted_dtypes = {}
    for output_name, inner_target_name in dtypes.items():
        converted_columns = {}
        for column_name, dtype in inner_target_name.items():
            dtype_name = dtype.name
            if dtype_name in dtype_mapping:
                primitive_type = dtype_mapping[dtype_name]
                converted_columns[column_name] = primitive_type
        converted_dtypes[output_name] = converted_columns

    return converted_dtypes


def gather_data_pointers_from_data_source(
    data_source: Path,
    validate: bool = True,
    output_inner_key: Optional[str] = None,
) -> dict[str, Path | int]:
    """
    Disk: ID -> file path
    Deeplake: ID -> integer index
    """
    iterator: (
        Generator[tuple[str, Path], None, None] | Generator[tuple[str, int], None, None]
    )
    if is_deeplake_dataset(data_source=str(data_source)):
        assert output_inner_key is not None
        iterator = build_deeplake_available_pointer_iterator(
            data_source=data_source,
            inner_key=output_inner_key,
        )
    else:
        iterator_base = get_file_path_iterator(
            data_source=data_source,
            validate=validate,
        )
        iterator = ((f.stem, f) for f in iterator_base)

    logger.debug("Gathering data pointers from %s.", data_source)
    id_to_pointer_mapping = {}
    for id_, pointer in tqdm(iterator, desc="Progress"):
        if id_ in id_to_pointer_mapping:
            raise ValueError(f"Duplicate ID: {id_}")

        id_to_pointer_mapping[id_] = pointer

    return id_to_pointer_mapping


def build_deeplake_available_pointer_iterator(
    data_source: Path, inner_key: str
) -> Generator[tuple[str, int], None, None]:
    deeplake_ds = load_deeplake_dataset(data_source=str(data_source))
    for int_pointer, sample in enumerate(deeplake_ds):
        inner_key_tensor = sample[inner_key]
        is_empty = inner_key_tensor.size == 0

        if is_empty:
            continue

        id_ = sample["ID"].numpy().item()

        yield id_, int(int_pointer)


def df_to_nested_dict(
    df: pd.DataFrame,
    dtypes: dict[str, dict[str, Any]],
    num_processes: int = -1,
) -> Dict:
    """
    Convert a DataFrame with a 2-level multi index ['ID', 'Output Name']
    to a nested dict.

    The resulting structure is

    {'ID': {output_name: {inner_output_identifier: inner_output_value}}}.

    For each output_name, only relevant columns are included.
    """

    if df.empty:
        return {}

    if num_processes == -1:
        dataset_size = len(df)
        if dataset_size < 10000:
            num_processes = 1
        else:
            cpu_count = os.cpu_count()
            if cpu_count is None:
                logger.warning("Could not determine number of CPUs. Using 1.")
                num_processes = 1
            else:
                num_processes = min(cpu_count, max(1, dataset_size // 10000))

    logger.debug(
        "Setting number of processes for parallel processing of DF to dict to %d.",
        num_processes,
    )

    chunk_size = max(1, len(df) // num_processes)
    chunks: list[pd.DataFrame] = []
    for i in range(0, len(df), chunk_size):
        start: int = i
        end: int = i + chunk_size
        chunks.append(df.iloc[start:end])

    process_func = partial(_process_chunk, dtypes=dtypes)
    with Pool(num_processes) as pool:
        dicts = pool.map(process_func, chunks)
    return _merge_dicts(dicts)


def _process_chunk(
    chunk: pd.DataFrame,
    dtypes: dict[str, dict[str, Any]],
) -> Dict[str, Dict[str, Dict[str, float | int | str]]]:
    nested_dict: Dict[str, Dict[str, Dict[str, float | int | str]]] = {}
    for row in chunk.itertuples(index=True, name="Row"):
        # Tuple of (ID, Output Name)
        multi_index = row.Index
        assert isinstance(multi_index, tuple)
        assert len(multi_index) == 2
        cur_id, cur_output_name = multi_index
        cur_output_dtypes = dtypes[cur_output_name]

        row_dict = {}
        # Skip the first element as it's the multi-index
        name_value_iter = zip(chunk.columns, row[1:])
        for column_name, value in name_value_iter:
            if pd.notna(value):
                dtype = cur_output_dtypes.get(column_name, None)
                if dtype:
                    value = dtype(value)
                row_dict[column_name] = value

        if cur_id not in nested_dict:
            nested_dict[cur_id] = {}
        nested_dict[cur_id][cur_output_name] = row_dict

    return nested_dict


def _merge_dicts(dicts: list[dict]) -> dict:
    final_dict = {}
    for d in dicts:
        for key, value in d.items():
            if key not in final_dict:
                final_dict[key] = value
            else:
                for output_name, output_values in value.items():
                    if output_name not in final_dict[key]:
                        final_dict[key][output_name] = output_values
                    else:
                        final_dict[key][output_name].update(output_values)
    return final_dict


def gather_all_ids_from_output_configs(
    output_configs: Sequence[schemas.OutputConfig],
) -> Tuple[str, ...]:
    all_ids: set[str] = set()
    for config in output_configs:
        cur_source = Path(config.output_info.output_source)
        if cur_source.suffix == ".csv":
            cur_ids = gather_ids_from_tabular_file(file_path=cur_source)
        elif cur_source.is_dir():
            cur_ids = gather_ids_from_data_source(data_source=cur_source)
        else:
            raise NotImplementedError(
                "Only csv and directory data sources are supported."
                f" Got: {cur_source}"
            )
        all_ids.update(cur_ids)

    return tuple(all_ids)


def read_manual_ids_if_exist(
    manual_valid_ids_file: Union[None, str]
) -> Union[Sequence[str], None]:
    if not manual_valid_ids_file:
        return None

    with open(manual_valid_ids_file, "r") as infile:
        manual_ids = tuple(line.strip() for line in infile)

    return manual_ids


def get_tabular_target_file_infos(
    output_configs: Iterable[schemas.OutputConfig],
) -> Dict[str, TabularFileInfo]:
    logger.debug("Setting up target labels.")

    tabular_files_info = {}

    for output_config in output_configs:
        if output_config.output_info.output_type != "tabular":
            continue

        output_name = output_config.output_info.output_name
        output_type_info = output_config.output_type_info
        assert isinstance(output_type_info, TabularOutputTypeConfig)

        tabular_info = TabularFileInfo(
            file_path=Path(output_config.output_info.output_source),
            con_columns=output_type_info.target_con_columns,
            cat_columns=output_type_info.target_cat_columns,
            parsing_chunk_size=output_type_info.label_parsing_chunk_size,
        )
        tabular_files_info[output_name] = tabular_info

    return tabular_files_info



from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, Type, cast

import torch
from torch import nn

from eir.models.model_setup_modules.meta_setup import (
    MetaClassGetterCallable,
    al_meta_model,
    get_default_meta_class,
    get_meta_model_class_and_kwargs_from_configs,
)
from eir.models.model_setup_modules.model_io import load_model
from eir.models.model_setup_modules.pretrained_setup import (
    overload_meta_model_feature_extractors_with_pretrained,
)
from eir.models.output.sequence.sequence_output_modules import (
    overload_embeddings_with_pretrained,
)
from eir.setup import schemas
from eir.setup.input_setup import al_input_objects_as_dict
from eir.train_utils.distributed import maybe_make_model_distributed
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.output_setup import al_output_objects_as_dict

al_model_registry = Dict[str, Callable[[str], Type[nn.Module]]]

logger = get_logger(name=__name__)


def get_model(
    global_config: schemas.GlobalConfig,
    inputs_as_dict: al_input_objects_as_dict,
    fusion_config: schemas.FusionConfig,
    outputs_as_dict: "al_output_objects_as_dict",
    meta_class_getter: MetaClassGetterCallable = get_default_meta_class,
) -> al_meta_model:
    meta_class, meta_kwargs = get_meta_model_class_and_kwargs_from_configs(
        global_config=global_config,
        fusion_config=fusion_config,
        inputs_as_dict=inputs_as_dict,
        outputs_as_dict=outputs_as_dict,
        meta_class_getter=meta_class_getter,
    )

    if global_config.m.pretrained_checkpoint:
        logger.info(
            "Loading pretrained checkpoint from '%s'.",
            global_config.m.pretrained_checkpoint,
        )
        loaded_meta_model = load_model(
            model_path=Path(global_config.m.pretrained_checkpoint),
            model_class=meta_class,
            model_init_kwargs=meta_kwargs,
            device=global_config.be.device,
            test_mode=False,
            strict_shapes=global_config.m.strict_pretrained_loading,
        )

        loaded_meta_model = overload_embeddings_with_pretrained(
            model=loaded_meta_model,
            inputs=inputs_as_dict,
            pretrained_checkpoint=global_config.m.pretrained_checkpoint,
        )

        return loaded_meta_model

    input_modules = overload_meta_model_feature_extractors_with_pretrained(
        input_modules=meta_kwargs["input_modules"],
        inputs_as_dict=inputs_as_dict,
        outputs_as_dict=outputs_as_dict,
        meta_class_getter=meta_class_getter,
    )
    meta_kwargs["input_modules"] = input_modules

    meta_model = meta_class(**meta_kwargs)
    device = torch.device(device=global_config.be.device)
    meta_model = meta_model.to(device=device)

    meta_model = maybe_make_model_distributed(
        device=global_config.be.device,
        model=meta_model,
    )

    compiled_model: al_meta_model
    if global_config.m.compile_model:
        compiled_model = cast(al_meta_model, torch.compile(model=meta_model))
    else:
        compiled_model = meta_model

    return compiled_model



from copy import copy
from typing import Any, Dict, Union

import timm
import torch
from timm.models import CoaT, CrossVit
from torch import nn

from eir.models.input.array.models_cnn import CNNModel, CNNModelConfig
from eir.models.input.image.image_models import (
    ImageModelClassGetterFunction,
    ImageModelConfig,
    ImageWrapperModel,
)
from eir.setup.input_setup_modules.common import DataDimensions
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def get_image_model(
    model_config: ImageModelConfig,
    data_dimensions: DataDimensions,
    model_registry_lookup: ImageModelClassGetterFunction,
    device: str,
) -> ImageWrapperModel:
    input_channels = data_dimensions.channels
    if model_config.model_type in _get_all_timm_models():
        feature_extractor = timm.create_model(
            model_name=model_config.model_type,
            pretrained=model_config.pretrained_model,
            num_classes=model_config.num_output_features,
            in_chans=input_channels,
        )

    elif model_config.model_type == "cnn":
        init_kwargs = _parse_init_kwargs(model_config=model_config)
        cnn_model_config = CNNModelConfig(**init_kwargs)

        feature_extractor = CNNModel(
            model_config=cnn_model_config,
            data_dimensions=data_dimensions,
        )

    else:
        assert isinstance(model_config.model_init_config, dict)
        feature_extractor = _meta_get_image_model_from_scratch(
            model_type=model_config.model_type,
            model_init_config=model_config.model_init_config,
            in_chans=input_channels,
            num_output_features=model_config.num_output_features,
        )

    feature_extractor = feature_extractor.to(device=device)

    _check_image_model_num_output_features_compatibility(
        feature_extractor=feature_extractor,
        num_output_features=model_config.num_output_features,
    )

    estimated_out_shape = estimate_feature_extractor_out_shape(
        feature_extractor=feature_extractor,
        data_dimensions=data_dimensions,
        num_output_features=model_config.num_output_features,
        device=device,
    )

    wrapper_model_class = model_registry_lookup(model_type="image-wrapper-default")
    model = wrapper_model_class(
        feature_extractor=feature_extractor,
        model_config=model_config,
        estimated_out_shape=estimated_out_shape,
    )

    if model_config.freeze_pretrained_model:
        for param in model.parameters():
            param.requires_grad = False

    return model


def _check_image_model_num_output_features_compatibility(
    feature_extractor: nn.Module, num_output_features: int
) -> None:
    match feature_extractor, num_output_features:
        case CrossVit(), 0:
            raise ValueError(
                "CrossVit model requires num_output_features to be set to a value "
                "greater than 0."
            )
        case CoaT(), 0:
            raise ValueError(
                "CoaT model requires num_output_features to be set to a value "
                "greater than 0."
            )


def estimate_feature_extractor_out_shape(
    feature_extractor: nn.Module,
    data_dimensions: DataDimensions,
    num_output_features: int,
    device: str,
) -> tuple[int, ...]:
    example_input = prepare_example_image_test_input(
        data_dimensions=data_dimensions,
        device=device,
    )

    has_forward_features = hasattr(feature_extractor, "forward_features")
    has_num_out_features = num_output_features > 0

    with torch.inference_mode():
        if has_forward_features and not has_num_out_features:
            feature_extractor_out = feature_extractor.forward_features(example_input)
        else:
            feature_extractor_out = feature_extractor(example_input)

    estimated_shape = feature_extractor_out.shape[1:]
    logger.debug("Estimated shape of feature extractor output: %s", estimated_shape)

    return estimated_shape


def prepare_example_image_test_input(
    data_dimensions: DataDimensions,
    device: str,
    batch_size: int = 2,
) -> torch.Tensor:
    full_shape = data_dimensions.full_shape()
    example_input = torch.rand((batch_size, *full_shape)).to(device=device)
    return example_input


def _get_all_timm_models() -> list[str]:
    base_models = timm.list_models()
    pretrained_models = timm.list_models(pretrained=True)

    all_models = base_models + pretrained_models
    all_models = list(set(all_models))
    all_models.sort()

    return all_models


def _parse_init_kwargs(model_config: ImageModelConfig) -> Dict[str, Any]:
    init_config = copy(model_config.model_init_config)
    if isinstance(init_config, dict):
        init_kwargs = init_config
    else:
        init_kwargs = init_config.__dict__
    assert isinstance(init_kwargs, dict)

    model_config_out_features = getattr(model_config, "num_output_features")
    init_kwargs_out_features = init_kwargs.get("num_output_features")
    if model_config_out_features:
        if model_config_out_features != init_kwargs_out_features:
            logger.warning(
                "num_output_features specified in model_config "
                "and model_init_config differ. "
                "Using value from model_config."
            )

            init_kwargs["num_output_features"] = model_config.num_output_features

    return init_kwargs


def _meta_get_image_model_from_scratch(
    model_type: str,
    model_init_config: dict,
    in_chans: int,
    num_output_features: Union[None, int],
) -> nn.Module:
    """
    A kind of ridiculous way to initialize modules from scratch that are found in timm,
    but could not find a better way at first glance.
    """

    feature_extractor_model_config = copy(model_init_config)
    feature_extractor_model_config["in_chans"] = in_chans
    if num_output_features:
        feature_extractor_model_config["num_classes"] = num_output_features

    logger.info(
        "Model '%s' not found among pretrained/external image model names, assuming "
        "module will be initialized from scratch using %s for initialization.",
        model_type,
        model_init_config,
    )

    feature_extractor_class = getattr(timm.models, model_type)
    parent_module = getattr(
        timm.models, feature_extractor_class.__module__.split(".")[-1]
    )
    found_modules = {
        k: getattr(parent_module, v)
        for k, v in feature_extractor_model_config.items()
        if isinstance(v, str) and getattr(parent_module, v, None)
    }
    feature_extractor_model_config_with_meta = {
        **feature_extractor_model_config,
        **found_modules,
    }

    feature_extractor = feature_extractor_class(
        **feature_extractor_model_config_with_meta
    )

    return feature_extractor



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Sequence

import numpy as np
import pandas as pd
import torch
from ignite.engine import Engine
from sklearn.preprocessing import LabelEncoder, StandardScaler

from eir.data_load.label_setup import al_label_transformers_object
from eir.models import model_training_utils
from eir.setup.output_setup import (
    ComputedArrayOutputInfo,
    ComputedImageOutputInfo,
    ComputedSequenceOutputInfo,
    ComputedTabularOutputInfo,
    al_output_objects_as_dict,
)
from eir.setup.schemas import GlobalConfig
from eir.target_setup.target_label_setup import MissingTargetsInfo
from eir.train_utils import metrics, utils
from eir.train_utils.evaluation_handlers.train_handlers_array_output import (
    array_out_single_sample_evaluation_wrapper,
)
from eir.train_utils.evaluation_handlers.train_handlers_sequence_output import (
    sequence_out_single_sample_evaluation_wrapper,
)
from eir.train_utils.latent_analysis import (
    LatentHookOutput,
    latent_analysis_wrapper,
    register_latent_hook,
)
from eir.utils.logging import get_logger
from eir.visualization import visualization_funcs as vf

if TYPE_CHECKING:
    from eir.train import Experiment
    from eir.train_utils.step_logic import al_training_labels_target
    from eir.train_utils.train_handlers import HandlerConfig

logger = get_logger(name=__name__, tqdm_compatible=True)


def validation_handler(engine: Engine, handler_config: "HandlerConfig") -> None:
    """
    A bit hacky how we manually attach metrics here, but that's because we
    don't want to evaluate as a running average (i.e. do it in the step
    function), but rather run over the whole validation dataset as we do
    in this function.
    """
    exp = handler_config.experiment
    gc = exp.configs.global_config
    iteration = engine.state.iteration

    exp.model.eval()

    hook_finalizers = register_pre_evaluation_hooks(model=exp.model, global_config=gc)

    output_generator = model_training_utils.get_prediction_outputs_generator(
        data_loader=exp.valid_loader,
        batch_prep_hook=exp.hooks.step_func_hooks.base_prepare_batch,
        batch_prep_hook_kwargs={"experiment": exp},
        model=exp.model,
        with_labels=True,
    )
    evaluation_results = run_split_evaluation(
        output_generator=output_generator,
        output_objects=exp.outputs,
        experiment_metrics=exp.metrics,
        loss_function=exp.loss_function,
        device=gc.be.device,
        missing_ids_per_output=exp.valid_dataset.missing_ids_per_output,
        with_labels=True,
    )

    hook_outputs = deregister_pre_evaluation_hooks(
        hook_finalizers=hook_finalizers,
        evaluation_results=evaluation_results,
    )

    run_all_eval_hook_analysis(
        hook_outputs=hook_outputs,
        run_folder=handler_config.run_folder,
        iteration=iteration,
    )

    write_eval_header = True if iteration == gc.ec.sample_interval else False
    metrics.persist_metrics(
        handler_config=handler_config,
        metrics_dict=evaluation_results.metrics_with_averages,
        iteration=iteration,
        write_header=write_eval_header,
        prefixes={"metrics": "validation_", "writer": "validation"},
    )

    if gc.ec.saved_result_detail_level >= 5:
        save_tabular_evaluation_results_wrapper(
            val_outputs=evaluation_results.gathered_outputs,
            val_labels=evaluation_results.gathered_labels,
            val_ids=evaluation_results.ids_per_output,
            iteration=iteration,
            experiment=handler_config.experiment,
        )

        sequence_out_single_sample_evaluation_wrapper(
            input_objects=exp.inputs,
            experiment=exp,
            iteration=iteration,
            auto_dataset_to_load_from=exp.valid_dataset,
            output_folder=gc.be.output_folder,
        )

        array_out_single_sample_evaluation_wrapper(
            input_objects=exp.inputs,
            experiment=exp,
            iteration=iteration,
            auto_dataset_to_load_from=exp.valid_dataset,
            output_folder=gc.be.output_folder,
        )

    exp.model.train()


def run_all_eval_hook_analysis(
    hook_outputs: dict[str, LatentHookOutput],
    run_folder: Path,
    iteration: int | str,
) -> None:
    for hook_name, hook_output in hook_outputs.items():
        match hook_output:
            case LatentHookOutput():
                latent_analysis_wrapper(
                    latent_outputs=hook_output,
                    run_folder=run_folder,
                    iteration=iteration,
                )
            case _:
                raise NotImplementedError()


def register_pre_evaluation_hooks(
    model: torch.nn.Module,
    global_config: GlobalConfig,
) -> dict[str, Callable]:
    model_hook_finalizers: dict[str, Callable] = {}

    if global_config.latent_sampling is not None:
        for layer_path in global_config.latent_sampling.layers_to_sample:
            model_hook_finalizers[f"latent_{layer_path}"] = register_latent_hook(
                model=model,
                layer_path=layer_path,
            )

    return model_hook_finalizers


def deregister_pre_evaluation_hooks(
    hook_finalizers: dict[str, Callable],
    evaluation_results: "EvaluationResults",
) -> dict[str, LatentHookOutput]:
    hook_outputs = {}
    for hook_name, hook_finalizer in hook_finalizers.items():
        hook_outputs[hook_name] = hook_finalizer(evaluation_results=evaluation_results)

    return hook_outputs


@dataclass()
class SplitModelOutputs:
    output_compute: dict[str, dict[str, torch.Tensor]]
    target_compute: dict[str, dict[str, torch.Tensor]]
    output_gather: dict[str, dict[str, torch.Tensor]]
    target_gather: dict[str, dict[str, torch.Tensor]]
    ids: list[str]


def get_split_output_generator(
    output_generator: Generator[
        model_training_utils.al_dataloader_gathered_predictions, None, None
    ],
    output_objects: al_output_objects_as_dict,
) -> Generator[SplitModelOutputs, None, None]:
    output_objects_by_type = split_output_objects_by_eval_type(
        output_objects=output_objects
    )

    for model_outputs, target_labels, ids in output_generator:
        output_gather = {
            k: v for k, v in model_outputs.items() if k in output_objects_by_type.gather
        }

        if target_labels is None:
            target_gather = {}
        else:
            target_gather = {
                k: v
                for k, v in target_labels.items()
                if k in output_objects_by_type.gather
            }

        output_compute = {
            k: v
            for k, v in model_outputs.items()
            if k in output_objects_by_type.compute
        }

        if target_labels is None:
            target_compute = {}
        else:
            target_compute = {
                k: v
                for k, v in target_labels.items()
                if k in output_objects_by_type.compute
            }

        yield SplitModelOutputs(
            output_compute=output_compute,
            target_compute=target_compute,
            output_gather=output_gather,
            target_gather=target_gather,
            ids=ids,
        )


@dataclass()
class EvaluationResults:
    metrics_with_averages: metrics.al_step_metric_dict
    gathered_outputs: dict[str, dict[str, torch.Tensor]]
    gathered_labels: dict[str, dict[str, torch.Tensor]]
    ids_per_output: dict[str, dict[str, list[str]]]
    all_ids: list[str]


def run_split_evaluation(
    output_generator: Generator[
        model_training_utils.al_dataloader_gathered_predictions, None, None
    ],
    output_objects: al_output_objects_as_dict,
    experiment_metrics: metrics.al_metric_record_dict,
    loss_function: Callable,
    device: str,
    missing_ids_per_output: MissingTargetsInfo,
    with_labels: bool = True,
) -> EvaluationResults:
    """
    This function is implemented in a way that allows for the evaluation to be
    split into two parts, evaluation that is performed:

        (1) On the full validation/predict dataset, for example to compute tabular
            metrics such as ROC-AUC and R2 that require the full dataset.
        (2) On a per-batch basis, then averaged across batches, for example
            loss for sequence outputs.

    It would be optimal to just use (1) for all output types, but this becomes
    a (GPU) RAM issue when the e.g. evaluation/prediction dataset is large,
    long sequences, and large embedding sizes are used.

    Note that applying (2) assumes that the same result is obtained when
    averaging across batches as when evaluating on the full dataset. Otherwise,
    the result will be an approximation, or possibly incorrect.
    """
    full_gathered_output_batches_total = []
    full_gathered_label_batches_total = []
    full_gathered_ids_total = []

    batch_computed_metrics = []
    batch_sizes = []

    split_output_objects = split_output_objects_by_eval_type(
        output_objects=output_objects
    )
    split_output_generator = get_split_output_generator(
        output_generator=output_generator, output_objects=output_objects
    )

    for split_outputs in split_output_generator:
        output_compute = split_outputs.output_compute
        target_compute = split_outputs.target_compute
        output_gather = split_outputs.output_gather
        target_gather = split_outputs.target_gather

        full_gathered_output_batches_total.append(output_gather)

        if with_labels:
            full_gathered_label_batches_total.append(target_gather)

        full_gathered_ids_total += split_outputs.ids

        if with_labels and len(split_output_objects.compute) > 0:
            cur_compute_metrics = compute_eval_metrics_wrapper(
                val_outputs=output_compute,
                val_target_labels=target_compute,
                outputs=split_output_objects.compute,
                experiment_metrics=experiment_metrics,
                loss_function=loss_function,
                device=device,
            )
            batch_computed_metrics.append(cur_compute_metrics)
            batch_sizes.append(len(split_outputs.ids))

    computed_metrics_averaged = average_nested_dict_values(
        data=batch_computed_metrics, weights=batch_sizes
    )

    stack_func = model_training_utils.stack_list_of_output_target_dicts
    full_gathered_label_batches_stacked = {}
    if with_labels:
        full_gathered_label_batches_stacked = stack_func(
            list_of_target_batch_dicts=full_gathered_label_batches_total
        )
    full_gathered_output_batches_stacked = stack_func(
        list_of_target_batch_dicts=full_gathered_output_batches_total
    )

    filtered_values = metrics.filter_missing_outputs_and_labels(
        batch_ids=full_gathered_ids_total,
        model_outputs=full_gathered_output_batches_stacked,
        target_labels=full_gathered_label_batches_stacked,
        missing_ids_info=missing_ids_per_output,
        with_labels=with_labels,
    )

    full_gathered_metrics = {}
    if with_labels and len(split_output_objects.gather) > 0:
        full_gathered_metrics = compute_eval_metrics_wrapper(
            val_outputs=filtered_values.model_outputs,
            val_target_labels=filtered_values.target_labels,
            outputs=split_output_objects.gather,
            experiment_metrics=experiment_metrics,
            loss_function=loss_function,
            device=device,
        )

    all_metrics = {**computed_metrics_averaged, **full_gathered_metrics}

    evaluation_results = EvaluationResults(
        metrics_with_averages=all_metrics,
        gathered_outputs=filtered_values.model_outputs,
        gathered_labels=filtered_values.target_labels,
        ids_per_output=filtered_values.ids,
        all_ids=full_gathered_ids_total,
    )

    return evaluation_results


def average_nested_dict_values(data: list[dict[str, Any]], weights: list[int]) -> dict:
    df = pd.json_normalize(data=data)

    weighted_sums = df.multiply(weights, axis=0).sum().to_dict()
    total_weight = sum(weights)

    weighted_averages_flat = {
        key: value / total_weight for key, value in weighted_sums.items()
    }
    averages_nested = unflatten_dict(flat_dict=weighted_averages_flat, separator=".")

    return averages_nested


def unflatten_dict(flat_dict: dict[str, Any], separator=".") -> dict[str, Any]:
    unflattened_dict: dict[str, Any] = {}
    for flat_key, value in flat_dict.items():
        keys = flat_key.split(separator)
        d = unflattened_dict
        for key in keys[:-1]:
            if key not in d:
                d[key] = {}
            d = d[key]
        d[keys[-1]] = value
    return unflattened_dict


@dataclass()
class SplitOutputObjectsByType:
    compute: al_output_objects_as_dict
    gather: al_output_objects_as_dict


def split_output_objects_by_eval_type(
    output_objects: al_output_objects_as_dict,
) -> SplitOutputObjectsByType:
    split_map = get_split_outputs_map(output_objects=output_objects)
    output_objects_compute = {
        k: i for k, i in output_objects.items() if split_map[k] == "compute"
    }
    output_objects_gather = {
        k: i for k, i in output_objects.items() if split_map[k] == "gather"
    }

    return SplitOutputObjectsByType(
        compute=output_objects_compute,
        gather=output_objects_gather,
    )


def get_split_outputs_map(output_objects: al_output_objects_as_dict) -> dict[str, str]:
    mapping = {}

    for output_name, output_object in output_objects.items():
        match output_object:
            case ComputedTabularOutputInfo():
                mapping[output_name] = "gather"
            case (
                ComputedSequenceOutputInfo()
                | ComputedArrayOutputInfo()
                | ComputedImageOutputInfo()
            ):
                mapping[output_name] = "compute"
            case _:
                raise ValueError(f"Unknown output type: {type(output_object)}")

    return mapping


def compute_eval_metrics_wrapper(
    val_outputs: dict[str, dict[str, torch.Tensor]],
    val_target_labels: "al_training_labels_target",
    outputs: al_output_objects_as_dict,
    experiment_metrics: metrics.al_metric_record_dict,
    loss_function: Callable,
    device: str,
) -> metrics.al_step_metric_dict:
    val_target_labels = model_training_utils.parse_tabular_target_labels(
        output_objects=outputs, device=device, labels=val_target_labels
    )

    eval_metrics_dict = metrics.calculate_batch_metrics(
        outputs_as_dict=outputs,
        outputs=val_outputs,
        labels=val_target_labels,
        mode="val",
        metric_record_dict=experiment_metrics,
    )

    val_losses = loss_function(inputs=val_outputs, targets=val_target_labels)
    val_loss_avg = metrics.aggregate_losses(losses_dict=val_losses)
    eval_metrics_dict_w_loss = metrics.add_loss_to_metrics(
        outputs_as_dict=outputs,
        losses=val_losses,
        metric_dict=eval_metrics_dict,
    )

    averaging_functions = experiment_metrics["averaging_functions"]
    assert isinstance(averaging_functions, dict)
    eval_metrics_dict_w_averages = metrics.add_multi_task_average_metrics(
        batch_metrics_dict=eval_metrics_dict_w_loss,
        outputs_as_dict=outputs,
        loss=val_loss_avg.item(),
        performance_average_functions=averaging_functions,
    )

    return eval_metrics_dict_w_averages


def save_tabular_evaluation_results_wrapper(
    val_outputs: Dict[str, Dict[str, torch.Tensor]],
    val_labels: Dict[str, Dict[str, torch.Tensor]],
    val_ids: dict[str, dict[str, list[str]]],
    iteration: int,
    experiment: "Experiment",
) -> None:
    for output_name, output_object in experiment.outputs.items():
        output_type = output_object.output_config.output_info.output_type
        if output_type != "tabular":
            continue

        assert isinstance(output_object, ComputedTabularOutputInfo)
        target_columns = output_object.target_columns
        for column_type, list_of_cols_of_this_type in target_columns.items():
            for column_name in list_of_cols_of_this_type:
                cur_sample_output_folder = utils.prepare_sample_output_folder(
                    output_folder=experiment.configs.gc.be.output_folder,
                    column_name=column_name,
                    output_name=output_name,
                    iteration=iteration,
                )
                cur_val_outputs = val_outputs[output_name][column_name]
                cur_val_outputs_np = cur_val_outputs.cpu().numpy()

                cur_val_labels = val_labels[output_name][column_name]
                cur_val_labels_np = cur_val_labels.cpu().numpy()
                target_transformers = output_object.target_transformers

                cur_val_ids = val_ids[output_name][column_name]

                plot_config = PerformancePlotConfig(
                    val_outputs=cur_val_outputs_np,
                    val_labels=cur_val_labels_np,
                    val_ids=cur_val_ids,
                    iteration=iteration,
                    column_name=column_name,
                    column_type=column_type,
                    target_transformer=target_transformers[column_name],
                    output_folder=cur_sample_output_folder,
                )

                save_evaluation_results(plot_config=plot_config)


@dataclass
class PerformancePlotConfig:
    val_outputs: np.ndarray
    val_labels: np.ndarray
    val_ids: List[str]
    iteration: int
    column_name: str
    column_type: str
    target_transformer: al_label_transformers_object
    output_folder: Path


def save_evaluation_results(
    plot_config: PerformancePlotConfig,
) -> None:
    pc = plot_config

    common_args = {
        "val_outputs": pc.val_outputs,
        "val_labels": pc.val_labels,
        "val_ids": pc.val_ids,
        "output_folder": pc.output_folder,
        "transformer": pc.target_transformer,
    }

    vf.gen_eval_graphs(plot_config=pc)

    if pc.column_type == "cat":
        save_classification_predictions(**common_args)
    elif pc.column_type == "con":
        scale_and_save_regression_predictions(**common_args)


def save_classification_predictions(
    val_labels: np.ndarray,
    val_outputs: np.ndarray,
    val_ids: List[str],
    transformer: LabelEncoder,
    output_folder: Path,
) -> None:
    validation_predictions_total = val_outputs.argmax(axis=1)

    df_predictions = _parse_valid_classification_predictions(
        val_true=val_labels,
        val_outputs=val_outputs,
        val_classes=transformer.classes_,
        val_predictions=validation_predictions_total,
        ids=np.array(val_ids),
    )

    df_predictions = _inverse_numerical_labels_hook(
        df=df_predictions, target_transformer=transformer
    )
    df_predictions.to_csv(output_folder / "predictions.csv", index=False)


def _parse_valid_classification_predictions(
    val_true: np.ndarray,
    val_predictions: np.ndarray,
    val_outputs: np.ndarray,
    val_classes: Sequence[str],
    ids: np.ndarray,
) -> pd.DataFrame:
    assert len(val_classes) == val_outputs.shape[1]

    columns = ["ID", "True_Label", "Predicted"]
    prediction_classes = [f"Score Class {i}" for i in val_classes]
    columns += prediction_classes

    column_values = [
        ids,
        val_true,
        val_predictions,
    ]

    for i in range(len(prediction_classes)):
        column_values.append(val_outputs[:, i])

    df = pd.DataFrame(columns=columns)

    for col_name, data in zip(columns, column_values):
        df[col_name] = data

    return df


def _inverse_numerical_labels_hook(
    df: pd.DataFrame, target_transformer: LabelEncoder
) -> pd.DataFrame:
    for column in ["True_Label", "Predicted"]:
        df[column] = target_transformer.inverse_transform(df[column])

    return df


def scale_and_save_regression_predictions(
    val_labels: np.ndarray,
    val_outputs: np.ndarray,
    val_ids: List[str],
    transformer: StandardScaler,
    output_folder: Path,
) -> None:
    val_labels_2d = val_labels.reshape(-1, 1)
    val_outputs_2d = val_outputs.reshape(-1, 1)

    val_labels = transformer.inverse_transform(val_labels_2d).squeeze()
    val_outputs = transformer.inverse_transform(val_outputs_2d).squeeze()

    data = np.array([val_ids, val_labels, val_outputs]).T
    df = pd.DataFrame(data=data, columns=["ID", "Actual", "Predicted"])
    df = df.set_index("ID")

    df.to_csv(output_folder / "regression_predictions.csv", index=True)
